from functools import lru_cache, total_ordering
import re
from typing import Union


class InvalidVersionFound(Exception):
    pass


@total_ordering
class Version:
    __slots__ = ['major', 'minor', 'patch']

    VERSION_RE = re.compile(r'^\s*(?P<major>\d+)\s*(?:\.\s*(?P<minor>\d+)\s*(?:\.\s*(?P<patch>\d+)\s*)?)?$')

    def __init__(
        self,
        major: Union[None, str, int] = None,
        minor: Union[None, str, int] = None,
        patch: Union[None, str, int] = None
    ):
        if isinstance(major, str) and not (minor or patch):  # Allow to pass in a str to directly create the object
            major = self._parse_version(major)

        if isinstance(major, Version):
            self.major = major.major
            self.minor = major.minor
            self.patch = major.patch
        else:
            self.major = int(major or 0)
            self.minor = int(minor or 0)
            self.patch = int(patch or 0)

    @staticmethod
    @lru_cache()  # Put this separately so I can cache the method. Small performance gain likely.
    def _parse_version(v):
        if v is None:
            return v

        if isinstance(v, Version):
            return v

        match = Version.VERSION_RE.match(v)
        if not match:
            raise InvalidVersionFound(f"Cannot parse version '{v}' (expected format: '1.2.3').")

        match = match.groupdict()

        return Version(
            major=int(match['major'] or 0),
            minor=int(match['minor'] or 0),
            patch=int(match['patch'] or 0),
        )

    def __eq__(self, other):
        if not isinstance(other, Version):
            other = Version(other)

        return (
            self.major == other.major
            and self.minor == other.minor
            and self.patch == other.patch
        )

    def __lt__(self, other):
        if not isinstance(other, Version):
            other = Version(other)

        return (
            (self.major, self.minor, self.patch)
            < (other.major, other.minor, other.patch)
        )

    def __hash__(self):
        return hash((self.major, self.minor, self.patch))

    def __repr__(self):
        return f"Version('{self.major}.{self.minor}.{self.patch}')"


class SparkVersion:
    attr_name = '_spark_version'

    def __get__(self, instance, owner):
        if not instance or not hasattr(instance, self.attr_name):
            return None

        return getattr(instance, self.attr_name)

    def __set__(self, instance, value):
        if value is None:
            self.__delete__(instance)
            return

        if not isinstance(value, Version):
            value = Version(value)

        setattr(instance, self.attr_name, value)

    def __delete__(self, instance):
        delattr(instance, self.attr_name)


class Configuration:
    """
    Configuration object to store dynamic changes in pysparklings behaviour.
    """

    spark_version = SparkVersion()


# Singleton object, only instantiated in this module
config = Configuration()



import datetime
from functools import partial
import re

from dateutil.tz import tzlocal
import pytz

from ._date_time import get_datetime_parser, get_time_formatter
from ._row import create_row
from .types import (
    ArrayType, BinaryType, BooleanType, ByteType, DateType, DecimalType, DoubleType, FloatType, IntegerType, LongType,
    MapType, NullType, NumericType, ShortType, StringType, StructType, TimestampType, UserDefinedType
)
from .utils import AnalysisException

TIME_REGEX = re.compile(
    "^([0-9]+):([0-9]+)?(?::([0-9]+))?(?:\\.([0-9]+))?(Z|[+-][0-9]+(?::(?:[0-9]+)?)?)?$"
)

GMT = pytz.timezone("GMT")


def identity(value, options):
    return value


def cast_from_none(value, from_type, options):
    if value is None:
        return None
    raise AnalysisException(
        f"Expected a null value from a field with type {from_type}, got {value}"
    )


def default_timestamp_formatter(timestamp):
    return timestamp.strftime("%Y-%m-%d %H:%M:%S")


def cast_to_string(value, from_type, options):
    date_format = get_time_formatter(options.get("dateformat", "yyyy-MM-dd"))
    timestamp_format = (get_time_formatter(options["timestampformat"])
                        if "timestampformat" in options
                        else default_timestamp_formatter)
    if value is None:
        return "null"
    if isinstance(from_type, DateType):
        return date_format(value)
    if isinstance(from_type, TimestampType):
        return timestamp_format(value)
    if isinstance(from_type, (ArrayType, StructType, MapType)):
        return cast_nested_to_str(value, from_type, options)
    if isinstance(from_type, BooleanType):
        return str(value).lower()
    return str(value)


def cast_nested_to_str(value, from_type, options):
    if isinstance(from_type, (ArrayType, StructType)):
        return cast_sequence(value, from_type, options)
    if isinstance(from_type, MapType):
        return cast_map(value, from_type, options)
    raise TypeError(f"Unable to cast {type(from_type)}")


def cast_map(value, from_type, options):
    casted_values = [
        (cast_to_string(key, from_type.keyType, options),
         (cast_to_string(sub_value, from_type.valueType, options)
          if sub_value is not None else None))
        for key, sub_value in value.items()
    ]
    return "[{0}]".format(
        ", ".join("{0} ->{1}".format(
            casted_key,
            f" {casted_value}" if casted_value is not None else ""
        ) for casted_key, casted_value in casted_values)
    )


def cast_sequence(value, from_type, options):
    if isinstance(from_type, StructType):
        types = [field.dataType for field in from_type.fields]
    else:
        types = [from_type.elementType] * len(value)
    casted_values = [
        cast_to_string(
            sub_value, sub_value_type, options
        ) if sub_value is not None else None
        for sub_value, sub_value_type in zip(value, types)
    ]

    return '[' + ','.join('' if x is None else f' {x}' for x in casted_values).strip() + ']'


def cast_to_binary(value, from_type, options):
    if isinstance(from_type, StringType):
        # noinspection PyTypeChecker
        return bytearray(value, 'utf-8')
    if isinstance(from_type, BinaryType):
        return value
    raise AnalysisException(f"Cannot cast type {from_type} to binary")


def cast_to_date(value, from_type, options):
    if isinstance(value, datetime.datetime):
        return value.date()
    if isinstance(value, datetime.date):
        return value
    if isinstance(value, str):
        # Spark cast only considers the first non empty part before a ' ' or a 'T'
        if ' ' in value:
            value = value.strip().split(" ")[0]
        if 'T' in value:
            value = value.split("T")[0]
        date_components = value.split("-")
        if len(date_components) > 3 or len(date_components[0]) != 4:
            return None
        # default month and day to 1
        date_components += ([1] * (3 - len(date_components)))
        try:
            return datetime.date(*map(int, date_components))
        except ValueError:
            return None
    if isinstance(from_type, (TimestampType, DateType, StringType)):
        return None  # other values would have been handle in the lines above

    raise AnalysisException(f"Cannot cast type {from_type} to date")


def parse_time_as_string(time_as_string):
    time_as_string = time_as_string.strip()
    match = TIME_REGEX.match(time_as_string)
    if not match:
        return None

    hour, minute, second, microsecond, tz_as_string = match.groups()
    if microsecond:
        microsecond = microsecond[:6]  # mimic Spark behaviour

    tzinfo = parse_timezone(tz_as_string)
    if tzinfo is None:  # Unable to parse
        return None

    return dict(
        hour=int(hour),
        minute=int(minute) if minute else 0,
        second=int(second) if second else 0,
        microsecond=int(microsecond) if microsecond else 0,
        tzinfo=tzinfo
    )


def parse_timezone(tz_as_string):
    if tz_as_string:
        if tz_as_string == "Z":
            tzinfo = GMT
        else:
            sign = tz_as_string[0]
            coeff = 1 if sign == "+" else -1
            if tz_as_string.count(":") == 1:
                hours, minutes = tz_as_string[1:].split(":")
                hours, minutes = int(hours), (int(minutes) if minutes else 0)
            elif tz_as_string.count(":") == 0:
                hours, minutes = int(tz_as_string[1:]), 0
            else:
                return None
            offset = coeff * datetime.timedelta(
                hours=hours,
                minutes=minutes
            )
            tzinfo = datetime.timezone(offset)
    else:
        tzinfo = tzlocal()
    return tzinfo


def cast_to_timestamp(value, from_type, options):
    if value == "" or value is None:
        return None
    if isinstance(value, str):
        date_as_string, time_as_string = split_datetime_as_string(value)
        date = cast_to_date(date_as_string, from_type, options=options)
        time_of_day = parse_time_as_string(time_as_string)

        return None if date is None or time_of_day is None else datetime.datetime(
            year=date.year,
            month=date.month,
            day=date.day,
            **time_of_day
        ).astimezone(tzlocal()).replace(tzinfo=None)
    if isinstance(value, datetime.datetime):
        return value
    if isinstance(value, datetime.date):
        return datetime.datetime(
            year=value.year,
            month=value.month,
            day=value.day
        )
    if isinstance(value, (int, float)):
        return datetime.datetime.fromtimestamp(value)
    if isinstance(from_type, (StringType, TimestampType, NumericType, BooleanType)):
        return None
    raise AnalysisException(f"Cannot cast type {from_type} to timestamp")


def split_datetime_as_string(value):
    first_space_position = (value.find(' ') + 1) or len(value)
    first_t_position = (value.find('T') + 1) or len(value)
    if first_space_position == len(value) and first_t_position == len(value):
        if ":" in value:
            # Value is only a time
            return datetime.date.today().strftime("%Y-%m-%d"), value
        # Value is only a date
        return value, "00:00:00"
    # Value is a datetime
    separation = min(first_space_position, first_t_position)
    date_as_string = value[:separation]
    time_as_string = value[separation:]
    return date_as_string, time_as_string


def cast_to_boolean(value, from_type, options):
    if value == "" or value is None:
        return None
    if isinstance(from_type, StringType):
        return True if value.lower() == "true" else False if value.lower() == "false" else None
    if isinstance(from_type, (NumericType, BooleanType)):
        return bool(value)
    raise AnalysisException(f"Cannot cast type {from_type} to boolean")


def _cast_to_bounded_type(name, min_value, max_value, value, from_type, options):
    if value == "" or value is None:
        return None
    size = max_value - min_value + 1
    if isinstance(from_type, DateType):
        return None
    if isinstance(from_type, TimestampType):
        return _cast_to_bounded_type(
            name,
            min_value,
            max_value,
            cast_to_float(value, from_type, options=options),
            FloatType(),
            options=options
        )
    if isinstance(from_type, StringType):
        casted_value = int(value)
        return casted_value if min_value <= casted_value <= max_value else None
    if isinstance(from_type, (NumericType, BooleanType)):
        value = int(value)
        return value % size if value % size <= max_value else value % -size
    raise AnalysisException(f"Cannot cast type {from_type} to {name}")


def cast_to_byte(value, from_type, options):
    min_value, max_value = -128, 127
    return _cast_to_bounded_type("byte", min_value, max_value, value, from_type, options=options)


def cast_to_short(value, from_type, options):
    min_value, max_value = -32768, 32767
    return _cast_to_bounded_type("short", min_value, max_value, value, from_type, options=options)


def cast_to_int(value, from_type, options):
    min_value, max_value = -2147483648, 2147483647
    return _cast_to_bounded_type("int", min_value, max_value, value, from_type, options=options)


def cast_to_long(value, from_type, options):
    min_value, max_value = -9223372036854775808, 9223372036854775807
    return _cast_to_bounded_type("long", min_value, max_value, value, from_type, options=options)


def cast_to_decimal(value, from_type, to_type, options):
    value_as_float = cast_to_float(value, from_type, options=options)
    if value_as_float is None:
        return None
    if value_as_float >= 10 ** (to_type.precision - to_type.scale):
        return None
    if to_type.scale == 0:
        return int(value_as_float)
    return round(value_as_float, ndigits=to_type.scale)


def cast_to_float(value, from_type, options):
    # NB: gelanis does not mimic the loss of accuracy of Spark nor value
    # bounding between float min&max values
    try:
        return cast_value(value, options=options)
    except ValueError as e:
        if isinstance(from_type, (DateType, TimestampType, NumericType, StringType)):
            return None
        raise AnalysisException(f"Cannot cast type {from_type} to float") from e


def cast_value(value, options):
    if value == "":
        return None
    if isinstance(value, datetime.datetime):
        return value.timestamp()
    if isinstance(value, datetime.date):
        return None
    if isinstance(value, (int, float)):
        return float(value)
    if isinstance(value, str):
        try:
            return float(value)
        except ValueError:
            return None
    raise ValueError("Unable to cast from value")


def cast_to_double(value, from_type, options):
    return cast_to_float(value, from_type, options=options)


def cast_to_array(value, from_type, to_type, options):
    if isinstance(from_type, ArrayType):
        caster = get_caster(
            from_type=from_type.elementType, to_type=to_type.elementType, options=options
        )
        return [
            caster(sub_value) if sub_value is not None else None
            for sub_value in value
        ]
    raise AnalysisException(f"Cannot cast type {from_type} to array")


def cast_to_map(value, from_type, to_type, options):
    if isinstance(from_type, MapType):
        key_caster = get_caster(
            from_type=from_type.keyType, to_type=to_type.keyType, options=options
        )
        value_caster = get_caster(
            from_type=from_type.valueType, to_type=to_type.valueType, options=options
        )
        return {
            key_caster(key): (value_caster(sub_value) if sub_value is not None else None)
            for key, sub_value in value.items()
        }
    raise AnalysisException(f"Cannot cast type {from_type} to map")


def cast_to_struct(value, from_type, to_type, options):
    if isinstance(from_type, StructType):
        return get_struct_caster(from_type, to_type, options)(value)
    raise NotImplementedError("Pysparkling does not support yet cast to struct")


def get_struct_caster(from_type, to_type, options):
    names = [to_field.name for to_field in to_type.fields]
    casters = [
        get_caster(from_field.dataType, to_field.dataType, options)
        for from_field, to_field in zip(from_type.fields, to_type.fields)
    ]

    def do_cast_to_struct(value):
        return create_row(
            names,
            (caster(sub_value) for caster, sub_value in zip(casters, value)),
            metadata=value.get_metadata()
        )

    return do_cast_to_struct


def cast_to_user_defined_type(value, from_type, options):
    raise NotImplementedError("Pysparkling does not support yet cast to UDF")


DESTINATION_DEPENDENT_CASTERS = {
    DecimalType: cast_to_decimal,
    ArrayType: cast_to_array,
    MapType: cast_to_map,
    StructType: cast_to_struct,
}

CASTERS = {
    StringType: cast_to_string,
    BinaryType: cast_to_binary,
    DateType: cast_to_date,
    TimestampType: cast_to_timestamp,
    # The ticket to expose CalendarIntervalType, in pyspark is SPARK-28492
    # It is open as this function is written, so we do not support it at the moment.
    # CalendarIntervalType: cast_to_interval,
    BooleanType: cast_to_boolean,
    ByteType: cast_to_byte,
    ShortType: cast_to_short,
    IntegerType: cast_to_int,
    FloatType: cast_to_float,
    LongType: cast_to_long,
    DoubleType: cast_to_double,
    UserDefinedType: cast_to_user_defined_type,
}


def get_caster(from_type, to_type, options):
    to_type_class = to_type.__class__
    if from_type == to_type:
        return partial(identity, options=options)
    if to_type_class == NullType:
        return partial(cast_from_none, from_type=from_type, options=options)
    if to_type_class == TimestampType:
        return get_datetime_parser(options.get("timestampFormat"))
    if to_type_class in DESTINATION_DEPENDENT_CASTERS:
        caster = DESTINATION_DEPENDENT_CASTERS[to_type_class]
        return partial(caster, from_type=from_type, to_type=to_type, options=options)
    if to_type_class in CASTERS:
        return partial(CASTERS[to_type_class], from_type=from_type, options=options)
    raise AnalysisException(f"Cannot cast from {from_type} to {to_type}")



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# This file is based on the PySpark version which in turn
# was ported from spark/util/StatCounter.scala
import copy
import math

try:
    from numpy import maximum, minimum, sqrt
except ImportError:
    maximum = max
    minimum = min
    sqrt = math.sqrt


class StatCounter:

    def __init__(self, values=None):
        self.n = 0  # Running count of our values
        self.mu = 0.0  # Running mean of our values
        self.m2 = 0.0  # Running variance numerator (sum of (x - mean)^2)
        self.maxValue = float("-inf")
        self.minValue = float("inf")

        if values:
            for v in values:
                self.merge(v)

    # Add a value into this StatCounter, updating the internal statistics.
    def merge(self, value):
        delta = value - self.mu
        self.n += 1
        self.mu += delta / self.n
        self.m2 += delta * (value - self.mu)
        self.maxValue = maximum(self.maxValue, value)
        self.minValue = minimum(self.minValue, value)

        return self

    # Merge another StatCounter into this one, adding up the
    # internal statistics.
    def mergeStats(self, other):
        if other is self:  # reference equality holds
            # Avoid overwriting fields in a weird order
            return self.mergeStats(other.copy())

        if self.n == 0:
            self.mu = other.mu
            self.m2 = other.m2
            self.n = other.n
            self.maxValue = other.maxValue
            self.minValue = other.minValue

        elif other.n != 0:
            delta = other.mu - self.mu
            if other.n * 10 < self.n:
                self.mu = self.mu + (delta * other.n) / (self.n + other.n)
            elif self.n * 10 < other.n:
                self.mu = other.mu - (delta * self.n) / (self.n + other.n)
            else:
                self.mu = ((self.mu * self.n + other.mu * other.n)
                           / (self.n + other.n))

            self.maxValue = maximum(self.maxValue, other.maxValue)
            self.minValue = minimum(self.minValue, other.minValue)

            self.m2 += other.m2 + ((delta * delta * self.n * other.n)
                                   / (self.n + other.n))
            self.n += other.n

        return self

    # Clone this StatCounter
    def copy(self):
        return copy.deepcopy(self)

    def count(self):
        return int(self.n)

    def mean(self):
        return self.mu

    def sum(self):
        return self.n * self.mu

    def min(self):
        return self.minValue

    def max(self):
        return self.maxValue

    # Return the variance of the values.
    def variance(self):
        if self.n == 0:
            return float('nan')

        return self.m2 / self.n

    #
    # Return the sample variance, which corrects for bias in estimating
    # the variance by dividing by N-1 instead of N.
    #
    def sampleVariance(self):
        if self.n <= 1:
            return float('nan')

        return self.m2 / (self.n - 1)

    # Return the standard deviation of the values.
    def stdev(self):
        return sqrt(self.variance())

    #
    # Return the sample standard deviation of the values, which corrects for
    # bias in estimating the variance by dividing by N-1 instead of N.
    #
    def sampleStdev(self):
        return sqrt(self.sampleVariance())

    def __repr__(self):
        return ("(count: %s, mean: %s, stdev: %s, max: %s, min: %s)" %
                (self.count(), self.mean(), self.stdev(), self.max(), self.min()))



from . import Expression
from ..types import StructField
from ..utils import AnalysisException


class FieldAsExpression(Expression):
    def __init__(self, field):
        super().__init__()
        self.field = field

    def eval(self, row, schema):
        return row[find_position_in_schema(schema, self.field)]

    def __str__(self):
        return self.field.name

    def output_fields(self, schema):
        return [self.field]

    def args(self):
        return (self.field,)


def find_position_in_schema(schema, expr):
    if isinstance(expr, str):
        show_id = False
        field_name = expr
        matches = set(i for i, field in enumerate(schema.fields) if field_name == field.name)
    elif isinstance(expr, FieldAsExpression):
        return find_position_in_schema(schema, expr.field)
    elif isinstance(expr, StructField) and hasattr(expr, "id"):
        show_id = True
        field_name = format_field(expr, show_id=show_id)
        matches = set(i for i, field in enumerate(schema.fields) if expr.id == field.id)
    else:
        if isinstance(expr, StructField):
            expression = f"Unbound field {expr.name}"
        else:
            expression = f"Expression type '{type(expr)}'"

        raise NotImplementedError(
            f"{expression} is not supported. "
            "As a user you should not see this error, feel free to report a bug at "
            "https://github.com/svenkreiss/gelanis/issues"
        )

    return get_checked_matches(matches, field_name, schema, show_id)


def get_checked_matches(matches, field_name, schema, show_id):
    if not matches:
        raise AnalysisException(f"Unable to find the column '{field_name}'"
                                f" among {format_schema(schema, show_id)}")

    if len(matches) > 1:
        raise AnalysisException(
            f"Reference '{field_name}' is ambiguous, found {len(matches)} columns matching it."
        )

    return matches.pop()


def format_schema(schema, show_id):
    return [format_field(field, show_id=show_id) for field in schema.fields]


def format_field(field, show_id):
    if show_id:
        return f"{field.name}#{field.id}"
    return field.name



from fnmatch import fnmatch
from io import BytesIO, StringIO
import logging

from ...exceptions import FileSystemNotSupported
from ...utils import parse_file_uri, Tokenizer
from .file_system import FileSystem

log = logging.getLogger(__name__)

try:
    from gcloud import storage
except ImportError:
    storage = None


class GS(FileSystem):
    """:class:`.FileSystem` implementation for Google Storage.

    Paths are of the form `gs://bucket_name/file_path` or
    `gs://project_name:bucket_name/file_path`.
    """

    #: Set a default project name.
    project_name = None

    #: Default mime type.
    mime_type = 'text/plain'

    _clients = {}

    def __init__(self, file_name):
        if storage is None:
            raise FileSystemNotSupported(
                'Google Storage is not supported. Install "gcloud".'
            )

        super().__init__(file_name)

        # obtain key
        t = Tokenizer(self.file_name)
        t.get_next('://')  # skip scheme
        bucket_name = t.get_next('/')
        if ':' in bucket_name:
            project_name, _, bucket_name = bucket_name.partition(':')
        else:
            project_name = GS.project_name
        blob_name = t.get_next()

        client = GS._get_client(project_name)
        bucket = client.get_bucket(bucket_name)
        self.blob = bucket.get_blob(blob_name)
        if not self.blob:
            self.blob = bucket.blob(blob_name)

    @staticmethod
    def _get_client(project_name):
        if project_name not in GS._clients:
            if storage is None:
                raise FileSystemNotSupported(
                    'Google Storage is not supported. Install "gcloud".'
                )
            GS._clients[project_name] = storage.Client(project_name)
        return GS._clients[project_name]

    @staticmethod
    def resolve_filenames(expr):
        files = []

        t = Tokenizer(expr)
        scheme = t.get_next('://')
        bucket_name = t.get_next('/')
        if ':' in bucket_name:
            project_name, _, bucket_name = bucket_name.partition(':')
        else:
            project_name = GS.project_name
        prefix = t.get_next(['*', '?'])

        bucket = GS._get_client(project_name).get_bucket(bucket_name)
        expr_s = len(scheme) + 3 + len(project_name) + 1 + len(bucket_name) + 1
        expr = expr[expr_s:]
        for k in bucket.list_blobs(prefix=prefix):
            if fnmatch(k.name, expr) or fnmatch(k.name, expr + '/part*'):
                files.append(f'{scheme}://{project_name}:{bucket_name}/{k.name}')
        return files

    @staticmethod
    def resolve_content(expr):
        scheme, raw_bucket_name, folder_path, pattern = parse_file_uri(expr)

        if ':' in raw_bucket_name:
            project_name, _, bucket_name = raw_bucket_name.partition(':')
        else:
            project_name = GS.project_name
            bucket_name = raw_bucket_name

        folder_path = folder_path[1:]  # Remove leading slash

        expr = f"{folder_path}{pattern}"
        # Match all files inside folders that match expr
        pattern_expr = f"{expr}{'' if expr.endswith('/') else '/'}*"

        bucket = GS._get_client(project_name).get_bucket(bucket_name)

        files = []
        for k in bucket.list_blobs(prefix=folder_path):
            if not k.name.endswith("/") and (
                    fnmatch(k.name, expr) or fnmatch(k.name, pattern_expr)
            ):
                files.append(
                    f'{scheme}://{raw_bucket_name}/{k.name}'
                )
        return files

    def exists(self):
        t = Tokenizer(self.file_name)
        t.get_next('//')  # skip scheme
        bucket_name = t.get_next('/')
        if ':' in bucket_name:
            project_name, _, bucket_name = bucket_name.partition(':')
        else:
            project_name = GS.project_name
        blob_name = t.get_next()
        bucket = GS._get_client(project_name).get_bucket(bucket_name)
        return (bucket.get_blob(blob_name)
                or list(bucket.list_blobs(prefix=f'{blob_name}/')))

    def load(self):
        log.debug('Loading %s with size %s.', self.blob.name, self.blob.size)
        return BytesIO(self.blob.download_as_string())

    def load_text(self, encoding='utf8', encoding_errors='ignore'):
        log.debug('Loading %s with size %s.', self.blob.name, self.blob.size)
        return StringIO(
            self.blob.download_as_string().decode(
                encoding, encoding_errors
            )
        )

    def dump(self, stream):
        log.debug('Dumping to %s.', self.blob.name)
        self.blob.upload_from_string(stream.read(),
                                     content_type=self.mime_type)
        return self

    def make_public(self, recursive=False):
        self.blob.make_public(recursive)
        return self



from .._casts import get_caster
from .._row import Row
from ..types import StructType
from .expressions import BinaryOperation, Expression, NullSafeBinaryOperation, TypeSafeBinaryOperation, UnaryExpression


class Negate(UnaryExpression):
    def eval(self, row, schema):
        return - self.column.eval(row, schema)

    def __str__(self):
        return f"(- {self.column})"


class Add(NullSafeBinaryOperation):
    def unsafe_operation(self, value1, value2):
        return value1 + value2

    def __str__(self):
        return f"({self.arg1} + {self.arg2})"


class Minus(NullSafeBinaryOperation):
    def unsafe_operation(self, value1, value2):
        return value1 - value2

    def __str__(self):
        return f"({self.arg1} - {self.arg2})"


class Time(NullSafeBinaryOperation):
    def unsafe_operation(self, value1, value2):
        return value1 * value2

    def __str__(self):
        return f"({self.arg1} * {self.arg2})"


class Divide(NullSafeBinaryOperation):
    def unsafe_operation(self, value1, value2):
        return value1 / value2 if value2 != 0 else None

    def __str__(self):
        return f"({self.arg1} / {self.arg2})"


class Mod(NullSafeBinaryOperation):
    def unsafe_operation(self, value1, value2):
        return value1 % value2

    def __str__(self):
        return f"({self.arg1} % {self.arg2})"


class Pow(NullSafeBinaryOperation):
    def unsafe_operation(self, value1, value2):
        return float(value1 ** value2)

    def __str__(self):
        return f"POWER({self.arg1}, {self.arg2})"


class Equal(TypeSafeBinaryOperation):
    def unsafe_operation(self, value_1, value_2):
        return value_1 == value_2

    def __str__(self):
        return f"({self.arg1} = {self.arg2})"


class LessThan(TypeSafeBinaryOperation):
    def unsafe_operation(self, value_1, value_2):
        return value_1 < value_2

    def __str__(self):
        return f"({self.arg1} < {self.arg2})"


class LessThanOrEqual(TypeSafeBinaryOperation):
    def unsafe_operation(self, value_1, value_2):
        return value_1 <= value_2

    def __str__(self):
        return f"({self.arg1} <= {self.arg2})"


class GreaterThan(TypeSafeBinaryOperation):
    def unsafe_operation(self, value_1, value_2):
        return value_1 > value_2

    def __str__(self):
        return f"({self.arg1} > {self.arg2})"


class GreaterThanOrEqual(TypeSafeBinaryOperation):
    def unsafe_operation(self, value_1, value_2):
        return value_1 >= value_2

    def __str__(self):
        return f"({self.arg1} >= {self.arg2})"


class And(TypeSafeBinaryOperation):
    def unsafe_operation(self, value_1, value_2):
        return value_1 and value_2

    def __str__(self):
        return f"({self.arg1} AND {self.arg2})"


class Or(TypeSafeBinaryOperation):
    def unsafe_operation(self, value_1, value_2):
        return value_1 or value_2

    def __str__(self):
        return f"({self.arg1} OR {self.arg2})"


class Invert(UnaryExpression):
    def eval(self, row, schema):
        value = self.column.eval(row, schema)
        if value is None:
            return None
        return not value

    def __str__(self):
        return f"(NOT {self.column})"


class BitwiseOr(BinaryOperation):
    def eval(self, row, schema):
        return self.arg1.eval(row, schema) | self.arg2.eval(row, schema)

    def __str__(self):
        return f"({self.arg1} | {self.arg2})"


class BitwiseAnd(BinaryOperation):
    def eval(self, row, schema):
        return self.arg1.eval(row, schema) & self.arg2.eval(row, schema)

    def __str__(self):
        return f"({self.arg1} & {self.arg2})"


class BitwiseXor(BinaryOperation):
    def eval(self, row, schema):
        return self.arg1.eval(row, schema) ^ self.arg2.eval(row, schema)

    def __str__(self):
        return f"({self.arg1} ^ {self.arg2})"


class BitwiseNot(UnaryExpression):
    def eval(self, row, schema):
        return ~(self.column.eval(row, schema))

    def __str__(self):
        return f"~{self.column}"


class EqNullSafe(BinaryOperation):
    def eval(self, row, schema):
        return self.arg1.eval(row, schema) == self.arg2.eval(row, schema)

    def __str__(self):
        return f"({self.arg1} <=> {self.arg2})"


class GetField(Expression):
    def __init__(self, item, field):
        super().__init__(item, field)
        self.item = item
        self.field = field

    def eval(self, row, schema):
        item_eval = self.item.eval(row, schema)
        if isinstance(item_eval, Row):
            item_value = dict(zip(
                item_eval.__fields__,
                item_eval
            ))
        elif isinstance(item_eval, dict):
            item_value = item_eval
        else:
            item_value = dict(enumerate(item_eval))
        field_value = self.field.eval(row, schema)
        return item_value.get(field_value)

    def __str__(self):
        if (hasattr(self.item.expr, "field")
                and hasattr(self.item.expr.field, "dataType")
                and isinstance(self.item.expr.field.dataType, StructType)):
            return f"{self.item}.{self.field}"
        return f"{self.item}[{self.field}]"

    def args(self):
        return (
            self.item,
            self.field
        )


class Contains(Expression):
    pretty_name = "contains"

    def __init__(self, expr, value):
        super().__init__(expr, value)
        self.expr = expr
        self.value = value

    def eval(self, row, schema):
        return self.value.eval(row, schema) in self.expr.eval(row, schema)

    def args(self):
        return (
            self.expr,
            self.value
        )


class StartsWith(Expression):
    pretty_name = "startswith"

    def __init__(self, arg1, substr):
        super().__init__(arg1, substr)
        self.arg1 = arg1
        self.substr = substr

    def eval(self, row, schema):
        return str(self.arg1.eval(row, schema)).startswith(self.substr)

    def args(self):
        return (
            self.arg1,
            self.substr
        )


class EndsWith(Expression):
    pretty_name = "endswith"

    def __init__(self, arg1, substr):
        super().__init__(arg1, substr)
        self.arg1 = arg1
        self.substr = substr

    def eval(self, row, schema):
        return str(self.arg1.eval(row, schema)).endswith(self.substr)

    def args(self):
        return (
            self.arg1,
            self.substr
        )


class IsIn(Expression):
    def __init__(self, arg1, cols):
        super().__init__(arg1)
        self.arg1 = arg1
        self.cols = [c.get_literal_value() for c in cols]

    def eval(self, row, schema):
        return self.arg1.eval(row, schema) in self.cols

    def __str__(self):
        all_cols = ', '.join(str(col) for col in self.cols)
        return f"({self.arg1} IN ({all_cols}))"

    def args(self):
        return [self.arg1] + self.cols


class IsNotNull(UnaryExpression):
    def eval(self, row, schema):
        return self.column.eval(row, schema) is not None

    def __str__(self):
        return f"({self.column} IS NOT NULL)"


class IsNull(UnaryExpression):
    def eval(self, row, schema):
        return self.column.eval(row, schema) is None

    def __str__(self):
        return f"({self.column} IS NULL)"


class Cast(Expression):
    def __init__(self, column, destination_type):
        super().__init__(column)
        self.column = column
        self.destination_type = destination_type
        self.caster = get_caster(
            from_type=self.column.data_type, to_type=destination_type, options={}
        )

    def eval(self, row, schema):
        return self.caster(self.column.eval(row, schema))

    def __str__(self):
        return str(self.column)

    def __repr__(self):
        return f"CAST({self.column} AS {self.destination_type.simpleString().upper()})"

    def args(self):
        return (
            self.column,
            self.destination_type
        )


class Substring(Expression):
    pretty_name = "substring"

    def __init__(self, expr, start, length):
        super().__init__(expr)
        self.expr = expr
        self.start = start.get_literal_value()
        self.length = length.get_literal_value()

    def eval(self, row, schema):
        return str(self.expr.eval(row, schema))[self.start - 1:self.start - 1 + self.length]

    def args(self):
        return (
            self.expr,
            self.start,
            self.length
        )


class Alias(Expression):
    def __init__(self, expr, alias: str):
        super().__init__(expr, alias)
        self.expr = expr
        self.alias = alias

    @property
    def may_output_multiple_cols(self):
        return self.expr.may_output_multiple_cols

    @property
    def may_output_multiple_rows(self):
        return self.expr.may_output_multiple_rows

    @property
    def is_an_aggregation(self):
        return self.expr.is_an_aggregation

    def eval(self, row, schema):
        return self.expr.eval(row, schema)

    def __str__(self):
        return self.alias

    def args(self):
        return (
            self.expr,
            self.alias
        )

    def data_type(self, schema):
        return self.expr.data_type(schema)


class UnaryPositive(UnaryExpression):
    def eval(self, row, schema):
        return self.column.eval(row, schema)

    def __str__(self):
        return f"(+ {self.column})"


__all__ = [
    "Negate",
    "Add",
    "Minus",
    "Time",
    "Divide",
    "Mod",
    "Pow",
    "Equal",
    "LessThan",
    "LessThanOrEqual",
    "GreaterThan",
    "GreaterThanOrEqual",
    "And",
    "Or",
    "Invert",
    "BitwiseOr",
    "BitwiseAnd",
    "BitwiseXor",
    "BitwiseNot",
    "EqNullSafe",
    "GetField",
    "Contains",
    "StartsWith",
    "EndsWith",
    "IsIn",
    "IsNotNull",
    "Cast",
    "Substring",
    "IsNull",
    "Alias",
    "UnaryPositive",
]



import logging
import operator
import pprint as py_pprint

from ..rdd import EmptyRDD

__all__ = ['DStream']

log = logging.getLogger(__name__)


class DStream:
    """A discrete stream of RDDs.

    Usually a DStream is created by a
    :class:`gelanis.streaming.StreamingContext` method like
    :func:`gelanis.streaming.StreamingContext.queueStream` and then
    operated on with the methods below.

    :param jdstream: previous stream
    :param StreamingContext ssc: the streaming context
    :param jrdd_deserializer: a deserializer callable
    """

    def __init__(self, jdstream, ssc, jrdd_deserializer=None):
        self._stream = jdstream
        self._context = ssc
        self._jrdd_deserializer = jrdd_deserializer

        self._current_time = 0.0
        self._current_rdd = None

        ssc._add_dstream(self)

    def _step(self, time_):
        if time_ <= self._current_time:
            return
        self._current_time = time_
        if self._jrdd_deserializer is None:
            self._current_rdd = self._stream.get()
        else:
            self._current_rdd = self._jrdd_deserializer(self._stream.get())

    def cache(self):
        """Cache RDDs.

        :rtype: DStream
        """
        return self.transform(lambda rdd: rdd.cache())

    def cogroup(self, other, numPartitions=None):
        """Apply cogroup to RDDs of this and other DStream.

        :param DStream other: another DStream
        :rtype: DStream


        Example:

        >>> import gelanis
        >>> sc = gelanis.Context()
        >>> ssc = gelanis.streaming.StreamingContext(sc, 0.1)
        >>> s1 = ssc.queueStream([[('a', 4), ('b', 2)], [('c', 7)]])
        >>> s2 = ssc.queueStream([[('a', 1), ('b', 3)], [('c', 8)]])
        >>> s1.cogroup(s2).foreachRDD(lambda rdd: print(sorted(rdd.collect())))
        >>> ssc.start()
        >>> ssc.awaitTermination(0.25)
        [('a', [[4], [1]]), ('b', [[2], [3]])]
        [('c', [[7], [8]])]
        """
        return CogroupedDStream(self, other, numPartitions)

    def context(self):
        """Return the StreamContext of this stream.

        :rtype: StreamingContext
        """
        return self._context

    def count(self):
        """Count elements per RDD.

        Creates a new RDD stream where each RDD has a single entry that
        is the count of the elements.

        :rtype: DStream
        """
        return (
            self
            .mapPartitions(lambda p: [sum(1 for _ in p)])
            .reduce(operator.add)
        )

    def countByValue(self):
        """Apply countByValue to every RDD.abs

        :rtype: DStream

        .. warning::
            Implemented as a local operation.


        Example:

        >>> import gelanis
        >>> sc = gelanis.Context()
        >>> ssc = gelanis.streaming.StreamingContext(sc, 0.1)
        >>> (
        ...     ssc
        ...     .queueStream([[1, 1, 5, 5, 5, 2]])
        ...     .countByValue()
        ...     .foreachRDD(lambda rdd: print(sorted(rdd.collect())))
        ... )
        >>> ssc.start()
        >>> ssc.awaitTermination(0.15)
        [(1, 2), (2, 1), (5, 3)]
        """
        return self.transform(
            lambda rdd: self._context._context.parallelize(
                rdd.countByValue().items()))

    def countByWindow(self, windowDuration, slideDuration=None):
        """Applies count() after window().

        :param float windowDuration: multiple of batch interval
        :param float slideDuration: multiple of batch interval
        :rtype: DStream


        Example:

        >>> import gelanis
        >>> sc = gelanis.Context()
        >>> ssc = gelanis.streaming.StreamingContext(sc, 0.1)
        >>> (
        ...     ssc
        ...     .queueStream([[1, 1, 5], [5, 5, 2, 4], [1, 2]])
        ...     .countByWindow(0.2)
        ...     .foreachRDD(lambda rdd: print(rdd.collect()))
        ... )
        >>> ssc.start()
        >>> ssc.awaitTermination(0.35)
        [3]
        [7]
        [6]
        """
        return self.window(windowDuration, slideDuration).count()

    def filter(self, f):
        """Filter elements.

        :param f: filter function
        :rtype: DStream
        """
        return self.transform(lambda rdd: rdd.filter(f))

    def flatMap(self, f, preservesPartitioning=False):
        """Apply function f and flatten.

        :param f: mapping function
        :rtype: DStream
        """
        return self.mapPartitions(
            lambda p: (e for pp in p for e in f(pp)),
            preservesPartitioning,
        )

    def flatMapValues(self, f):
        """Apply f to each value of a key-value pair.

        :param f: map function
        :rtype: DStream
        """
        return self.transform(lambda rdd: rdd.flatMapValues(f))

    def foreachRDD(self, func):
        """Apply func.

        :param func: Function to apply.
        """
        self.transform(func)

    def fullOuterJoin(self, other, numPartitions=None):
        """Apply fullOuterJoin to each pair of RDDs.

        :rtype: DStream


        Example:

        >>> import gelanis
        >>> sc = gelanis.Context()
        >>> ssc = gelanis.streaming.StreamingContext(sc, 0.1)
        >>> s1 = ssc.queueStream([[('a', 4), ('b', 2)], [('c', 7)]])
        >>> s2 = ssc.queueStream([[('a', 1), ('b', 3)], [('c', 8)]])
        >>> (
        ...     s1.fullOuterJoin(s2)
        ...     .foreachRDD(lambda rdd: print(sorted(rdd.collect())))
        ... )
        >>> ssc.start()
        >>> ssc.awaitTermination(0.25)
        [('a', (4, 1)), ('b', (2, 3))]
        [('c', (7, 8))]


        Example with repeated keys:

        >>> import gelanis
        >>> sc = gelanis.Context()
        >>> ssc = gelanis.streaming.StreamingContext(sc, 0.1)
        >>> s1 = ssc.queueStream([[('a', 4), ('a', 2)], [('c', 7)]])
        >>> s2 = ssc.queueStream([[('b', 1)], [('c', 8)]])
        >>> (
        ...     s1.fullOuterJoin(s2)
        ...     .foreachRDD(lambda rdd: print(sorted(rdd.collect())))
        ... )
        >>> ssc.start()
        >>> ssc.awaitTermination(0.25)
        [('a', (2, None)), ('a', (4, None)), ('b', (None, 1))]
        [('c', (7, 8))]
        """
        return CogroupedDStream(self, other, numPartitions, op='fullOuterJoin')

    def groupByKey(self):
        """group by key

        :rtype: DStream
        """
        return self.transform(lambda rdd: rdd.groupByKey())

    def join(self, other, numPartitions=None):
        """Apply join to each pair of RDDs.

        :rtype: DStream


        Example:

        >>> import gelanis
        >>> sc = gelanis.Context()
        >>> ssc = gelanis.streaming.StreamingContext(sc, 0.1)
        >>> s1 = ssc.queueStream([[('a', 4), ('e', 2)], [('c', 7)]])
        >>> s2 = ssc.queueStream([[('a', 1), ('b', 3)], [('c', 8)]])
        >>> (
        ...     s1.join(s2)
        ...     .foreachRDD(lambda rdd: print(sorted(rdd.collect())))
        ... )
        >>> ssc.start()
        >>> ssc.awaitTermination(0.25)
        [('a', (4, 1))]
        [('c', (7, 8))]
        """
        return CogroupedDStream(self, other, numPartitions, op='join')

    def leftOuterJoin(self, other, numPartitions=None):
        """Apply leftOuterJoin to each pair of RDDs.

        :rtype: DStream


        Example:

        >>> import gelanis
        >>> sc = gelanis.Context()
        >>> ssc = gelanis.streaming.StreamingContext(sc, 0.1)
        >>> s1 = ssc.queueStream([[('a', 4), ('e', 2)], [('c', 7)]])
        >>> s2 = ssc.queueStream([[('a', 1), ('b', 3)], [('c', 8)]])
        >>> (
        ...     s1.leftOuterJoin(s2)
        ...     .foreachRDD(lambda rdd: print(sorted(rdd.collect())))
        ... )
        >>> ssc.start()
        >>> ssc.awaitTermination(0.25)
        [('a', (4, 1)), ('e', (2, None))]
        [('c', (7, 8))]
        """
        return CogroupedDStream(self, other, numPartitions, op='leftOuterJoin')

    def map(self, f, preservesPartitioning=False):
        """Apply function f

        :param f: mapping function
        :rtype: DStream


        Example:

        >>> import gelanis
        >>> sc = gelanis.Context()
        >>> ssc = gelanis.streaming.StreamingContext(sc, 0.1)
        >>> (
        ...     ssc
        ...     .queueStream([[4], [2], [7]])
        ...     .map(lambda e: e + 1)
        ...     .foreachRDD(lambda rdd: print(rdd.collect()))
        ... )
        >>> ssc.start()
        >>> ssc.awaitTermination(0.35)
        [5]
        [3]
        [8]
        """
        return (
            self
            .mapPartitions(lambda p: (f(e) for e in p), preservesPartitioning)
            .transform(lambda rdd:
                       rdd.setName(f'{rdd.prev.name()}:{f}'))
        )

    def mapPartitions(self, f, preservesPartitioning=False):
        """Map partitions.

        :param f: mapping function
        :rtype: DStream
        """
        return (
            self
            .mapPartitionsWithIndex(lambda i, p: f(p), preservesPartitioning)
            .transform(lambda rdd:
                       rdd.setName(f'{rdd.prev.name()}:{f}'))
        )

    def mapPartitionsWithIndex(self, f, preservesPartitioning=False):
        """Apply a map function that takes an index and the data.

        Map partitions with a function that takes the partition index
        and an iterator over the partition data as arguments.

        :param f: mapping function
        :rtype: DStream
        """
        return self.transform(
            lambda rdd: rdd.mapPartitionsWithIndex(f, preservesPartitioning)
        )

    def mapValues(self, f):
        """Apply ``f`` to every element.

        :rtype: DStream


        Example:

        >>> import gelanis
        >>> sc = gelanis.Context()
        >>> ssc = gelanis.streaming.StreamingContext(sc, 0.1)
        >>> (
        ...     ssc
        ...     .queueStream([[('a', 4)], [('b', 2)], [('c', 7)]])
        ...     .mapValues(lambda e: e + 1)
        ...     .foreachRDD(lambda rdd: print(rdd.collect()))
        ... )
        >>> ssc.start()
        >>> ssc.awaitTermination(0.35)
        [('a', 5)]
        [('b', 3)]
        [('c', 8)]
        """
        return self.transform(lambda rdd: rdd.mapValues(f))

    def pprint(self, num=10):
        """Print the first ``num`` elements of each RDD.

        :param int num: Set number of elements to be printed.
        """

        def pprint_map(time_, rdd):
            print(f'>>> Time: {time_}')
            data = rdd.take(num + 1)
            for d in data[:num]:
                py_pprint.pprint(d)
            if len(data) > num:
                print('...')
            print('')

        self.foreachRDD(pprint_map)

    def reduce(self, func):
        """Return a new DStream where each RDD was reduced with ``func``.

        :rtype: DStream
        """

        # avoid RDD.reduce() which does not return an RDD
        return self.transform(
            lambda rdd: (
                rdd
                .map(lambda i: (None, i))
                .reduceByKey(func)
                .map(lambda none_i: none_i[1])
            )
        )

    def reduceByKey(self, func, numPartitions=None):
        """Apply reduceByKey to every RDD.

        :param func: reduce function to apply
        :param int numPartitions: number of partitions
        :rtype: DStream
        """
        return self.transform(lambda rdd: rdd.reduceByKey(func))

    def repartition(self, numPartitions):
        """Repartition every RDD.

        :rtype: DStream


        Example:

        >>> import gelanis
        >>> sc = gelanis.Context()
        >>> ssc = gelanis.streaming.StreamingContext(sc, 0.1)
        >>> (
        ...     ssc
        ...     .queueStream([['hello', 'world']])
        ...     .repartition(2)
        ...     .foreachRDD(lambda rdd: print(len(rdd.partitions())))
        ... )
        >>> ssc.start()
        >>> ssc.awaitTermination(0.25)
        2
        0

        """
        return self.transform(
            lambda rdd: (rdd.repartition(numPartitions)
                         if not isinstance(rdd, EmptyRDD) else rdd)
        )

    def rightOuterJoin(self, other, numPartitions=None):
        """Apply rightOuterJoin to each pair of RDDs.

        :rtype: DStream


        Example:

        >>> import gelanis
        >>> sc = gelanis.Context()
        >>> ssc = gelanis.streaming.StreamingContext(sc, 0.1)
        >>> s1 = ssc.queueStream([[('a', 4), ('e', 2)], [('c', 7)]])
        >>> s2 = ssc.queueStream([[('a', 1), ('b', 3)], [('c', 8)]])
        >>> (
        ...     s1.rightOuterJoin(s2)
        ...     .foreachRDD(lambda rdd: print(sorted(rdd.collect())))
        ... )
        >>> ssc.start()
        >>> ssc.awaitTermination(0.25)
        [('a', (4, 1)), ('b', (None, 3))]
        [('c', (7, 8))]
        """
        return CogroupedDStream(self, other, numPartitions,
                                op='rightOuterJoin')

    def saveAsTextFiles(self, prefix, suffix=None):
        """Save every RDD as a text file (or sets of text files).

        :param string prefix: path prefix of the output
        :param string suffix: file suffix (e.g. '.gz' to enable compression)


        Example:

        >>> from backports import tempfile
        >>> import os, gelanis
        >>> sc = gelanis.Context()
        >>> ssc = gelanis.streaming.StreamingContext(sc, 0.1)
        >>> with tempfile.TemporaryDirectory() as tmp_dir:
        ...     (
        ...         ssc.queueStream([['hello', 'world'], [1, 2]])
        ...         .saveAsTextFiles(os.path.join(tmp_dir, 'textout'))
        ...     )
        ...     ssc.start()
        ...     ssc.awaitTermination(0.25)
        ...     result = sc.textFile(tmp_dir + '*').collect()
        >>> result
        ['hello', 'world', '1', '2']
        """
        self.foreachRDD(
            lambda time_, rdd:
            rdd.saveAsTextFile(f'{prefix}-{time_ * 1000:.0f}{suffix if suffix is not None else ""}')
        )

    def slice(self, begin, end):
        """Filter RDDs to between begin and end.

        :param datetime.datetime|int begin: datetiem or unix timestamp
        :param datetime.datetime|int end: datetiem or unix timestamp
        :rtype: DStream
        """
        return self.transform(lambda time_, rdd:
                              rdd if begin <= time_ <= end
                              else EmptyRDD(self._context._context))

    def transform(self, func):
        """Return a new DStream where each RDD is transformed by ``f``.

        :param f: Function that transforms an RDD.
        :rtype: DStream
        """
        if func.__code__.co_argcount == 1:
            one_arg_func = func
            func = lambda _, rdd: one_arg_func(rdd)

        return TransformedDStream(self, func)

    def transformWith(self, func, other, keepSerializer=False):
        """Return a new DStream where each RDD is transformed by ``f``.

        :param f: transformation function
        :rtype: DStream

        The transformation function can have arguments ``(time, rdd_a, rdd_b)``
        or ``(rdd_a, rdd_b)``.
        """
        if func.__code__.co_argcount == 2:
            two_arg_func = func
            func = lambda _, rdd_a, rdd_b: two_arg_func(rdd_a, rdd_b)

        return TransformedWithDStream(self, func, other)

    def union(self, other):
        """Union of two DStreams.

        :param DStream other: Another DStream.


        Example:

        >>> import gelanis
        >>> sc = gelanis.Context()
        >>> ssc = gelanis.streaming.StreamingContext(sc, 0.1)
        >>> odd = ssc.queueStream([[1], [3], [5]])
        >>> even = ssc.queueStream([[2], [4], [6]])
        >>> (
        ...     odd.union(even)
        ...     .foreachRDD(lambda rdd: print(rdd.collect()))
        ... )
        >>> ssc.start()
        >>> ssc.awaitTermination(0.35)
        [1, 2]
        [3, 4]
        [5, 6]
        """
        def union_rdds(rdd_a, rdd_b):
            return self._context._context.union((rdd_a, rdd_b))

        return self.transformWith(union_rdds, other)

    def updateStateByKey(self, func):
        """Process with state.

        :param func: Evaluated per key. Takes list of input_values and a state.
        :rtype: DStream


        This example shows how to return the latest value per key:

        >>> import gelanis
        >>> sc = gelanis.Context()
        >>> ssc = gelanis.streaming.StreamingContext(sc, 0.2)
        >>> (
        ...     ssc
        ...     .queueStream([[('a', 1), ('b', 3)], [('a', 2), ('c', 4)]])
        ...     .updateStateByKey(lambda input_values, state:
        ...                       state
        ...                       if not input_values
        ...                       else input_values[-1])
        ...     .foreachRDD(lambda rdd: print(sorted(rdd.collect())))
        ... )
        >>> ssc.start()
        >>> ssc.awaitTermination(0.5)
        [('a', 1), ('b', 3)]
        [('a', 2), ('b', 3), ('c', 4)]


        This example counts values per key:

        >>> sc = gelanis.Context()
        >>> ssc = gelanis.streaming.StreamingContext(sc, 0.2)
        >>> (
        ...     ssc
        ...     .queueStream([[('a', 1)], [('a', 2), ('b', 4), ('b', 3)]])
        ...     .updateStateByKey(lambda input_values, state:
        ...                       (state if state is not None else 0) +
        ...                       sum(input_values))
        ...     .foreachRDD(lambda rdd: print(sorted(rdd.collect())))
        ... )
        >>> ssc.start()
        >>> ssc.awaitTermination(0.5)
        [('a', 1)]
        [('a', 3), ('b', 7)]
        """
        return StatefulDStream(self, func)

    def window(self, windowDuration, slideDuration=None):
        """Windowed RDD.

        :param float windowDuration: multiple of batching interval
        :param float slideDuration: multiple of batching interval
        :rtype: DStream


        Example:

        >>> import gelanis
        >>> sc = gelanis.Context()
        >>> ssc = gelanis.streaming.StreamingContext(sc, 0.2)
        >>> (
        ...     ssc
        ...     .queueStream([[1], [2], [3], [4], [5], [6]])
        ...     .window(0.6)
        ...     .foreachRDD(lambda rdd: print(rdd.collect()))
        ... )
        >>> ssc.start()
        >>> ssc.awaitTermination(1.3)
        [1]
        [1, 2]
        [1, 2, 3]
        [2, 3, 4]
        [3, 4, 5]
        [4, 5, 6]
        """
        return WindowedDStream(self, windowDuration, slideDuration)


class TransformedDStream(DStream):
    def __init__(self, prev, func):
        super().__init__(prev._stream, prev._context)
        self._prev = prev
        self._func = func

    def _step(self, time_):
        if time_ <= self._current_time:
            return

        self._prev._step(time_)
        self._current_time = time_
        self._current_rdd = self._func(time_, self._prev._current_rdd)


class TransformedWithDStream(DStream):
    def __init__(self, prev, func, other_prev):
        super().__init__(
            prev._stream, prev._context)
        self._prev = prev
        self._func = func
        self._other_prev = other_prev

    def _step(self, time_):
        if time_ <= self._current_time:
            return

        self._prev._step(time_)
        self._other_prev._step(time_)
        self._current_time = time_
        self._current_rdd = self._func(time_,
                                       self._prev._current_rdd,
                                       self._other_prev._current_rdd)


class WindowedDStream(DStream):
    def __init__(self, prev, windowDuration, slideDuration=None):
        super().__init__(prev._stream, prev._context)

        if slideDuration is None:
            slideDuration = self._context.batch_duration

        self._prev = prev
        self._window_duration = int(round(
            windowDuration / self._context.batch_duration))
        self._slide_duration = int(round(
            slideDuration / self._context.batch_duration))
        self._slide_counter = 0
        self._window = []

    def _step(self, time_):
        if time_ <= self._current_time:
            return

        self._prev._step(time_)
        self._window.append(self._prev._current_rdd)

        # window duration
        while len(self._window) > self._window_duration:
            self._window.pop(0)

        # slide duration
        self._slide_counter = (self._slide_counter + 1) % self._slide_duration
        if self._slide_counter != 0:
            return

        self._current_time = time_
        self._current_rdd = self._context._context.union(self._window)


class CogroupedDStream(DStream):
    def __init__(self, prev1, prev2, numPartitions=None, op='cogroup'):
        super().__init__(prev1._stream, prev1._context)
        self._prev1 = prev1
        self._prev2 = prev2
        self._num_partitions = numPartitions
        self._op = op

    def _step(self, time_):
        if time_ <= self._current_time:
            return

        self._prev1._step(time_)
        self._prev2._step(time_)
        self._current_time = time_
        self._current_rdd = getattr(self._prev1._current_rdd, self._op)(
            self._prev2._current_rdd, self._num_partitions)


class StatefulDStream(DStream):
    def __init__(self, prev, state_update_fn):
        super().__init__(prev._stream, prev._context)
        self._prev = prev
        self._func = state_update_fn
        self._state_rdd = EmptyRDD(self._context._context)

    def convert_fn(self, joined):
        input_values, state_list = joined
        state = state_list[-1] if state_list else None

        return self._func(input_values, state)

    def _step(self, time_):
        if time_ <= self._current_time:
            return

        self._prev._step(time_)
        self._current_time = time_

        combined = self._prev._current_rdd.cogroup(self._state_rdd)
        self._state_rdd = combined.mapValues(self.convert_fn)
        self._current_rdd = self._state_rdd



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from array import array
from collections import ChainMap
import ctypes
import datetime
import decimal
import itertools
import json as _json
import re
import sys
from typing import Union

from ._row import create_row, Row
from .utils import ParseException

__all__ = [
    "DataType", "NullType", "StringType", "BinaryType", "BooleanType", "DateType",
    "TimestampType", "DecimalType", "DoubleType", "FloatType", "ByteType", "IntegerType",
    "LongType", "ShortType", "ArrayType", "MapType", "StructField", "StructType"
]


class DataType:
    """Base class for data types."""

    def __repr__(self):
        return self.__class__.__name__

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and set(self.__dict__.keys()) == set(other.__dict__.keys())
            and all(self.__dict__[k] == other.__dict__[k] for k in self.__dict__)
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    @classmethod
    def typeName(cls):
        return cls.__name__[:-4].lower()

    def simpleString(self):
        return self.typeName()

    def jsonValue(self):
        return self.typeName()

    def json(self):
        return _json.dumps(self.jsonValue(),
                           separators=(',', ':'),
                           sort_keys=True)

    def needConversion(self):
        """
        Return whether or not value of this type is converted by Spark
        between DF definition and internal SQL object.

        This is used to avoid the unnecessary conversion for ArrayType/MapType/StructType.
        """
        return False

    def toInternal(self, obj):
        """
        This method is not used by gelanis

        Converts a Python object into an internal SQL object.
        """
        return obj

    def fromInternal(self, obj):
        """
        This method is not used by gelanis

        Converts an internal SQL object into a native Python object.
        """
        return obj


# This singleton pattern does not work with pickle, you will get
# another object after pickle and unpickle
class DataTypeSingleton(type):
    """Metaclass for DataType"""

    _instances = {}

    def __call__(cls):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__()
        return cls._instances[cls]


class NullType(DataType, metaclass=DataTypeSingleton):
    """Null type.

    The data type representing None, used for the types that cannot be inferred.
    """


class AtomicType(DataType):
    """An internal type used to represent everything that is not
    null, UDTs, arrays, structs, and maps."""


class NumericType(AtomicType):
    """Numeric data types.
    """


class IntegralType(NumericType, metaclass=DataTypeSingleton):
    """Integral data types.
    """


class FractionalType(NumericType):
    """Fractional data types.
    """


class StringType(AtomicType, metaclass=DataTypeSingleton):
    """String data type.
    """


class BinaryType(AtomicType, metaclass=DataTypeSingleton):
    """Binary (byte array) data type."""


class BooleanType(AtomicType, metaclass=DataTypeSingleton):
    """Boolean data type."""


class DateType(AtomicType, metaclass=DataTypeSingleton):
    """Date (datetime.date) data type."""

    EPOCH_ORDINAL = datetime.datetime(1970, 1, 1).toordinal()

    def needConversion(self):
        return False


class TimestampType(AtomicType, metaclass=DataTypeSingleton):
    """Timestamp (datetime.datetime) data type."""

    def needConversion(self):
        return True

    def toInternal(self, obj):
        if obj.tzinfo is not None:
            return obj.astimezone()
        return obj


class DecimalType(FractionalType):
    """Decimal (decimal.Decimal) data type.

    The DecimalType must have fixed precision (the maximum total number of digits)
    and scale (the number of digits on the right of dot). For example, (5, 2) can
    support the value from [-999.99 to 999.99].

    The precision can be up to 38, the scale must be less or equal to precision.

    When create a DecimalType, the default precision and scale is (10, 0). When infer
    schema from decimal.Decimal objects, it will be DecimalType(38, 18).

    :param precision: the maximum total number of digits (default: 10)
    :param scale: the number of digits on right side of dot. (default: 0)
    """

    def __init__(self, precision=10, scale=0):
        self.precision = int(precision)
        self.scale = int(scale)
        self.hasPrecisionInfo = True  # this is public API

    def simpleString(self):
        return "decimal(%d,%d)" % (self.precision, self.scale)

    def jsonValue(self):
        return "decimal(%d,%d)" % (self.precision, self.scale)

    def __repr__(self):
        return "DecimalType(%d,%d)" % (self.precision, self.scale)


class DoubleType(FractionalType, metaclass=DataTypeSingleton):
    """Double data type, representing double precision floats."""


class FloatType(FractionalType, metaclass=DataTypeSingleton):
    """Float data type, representing single precision floats."""


class ByteType(IntegralType):
    """Byte data type, i.e. a signed integer in a single byte.
    """

    def simpleString(self):
        return 'tinyint'


class IntegerType(IntegralType):
    """Int data type, i.e. a signed 32-bit integer.
    """

    def simpleString(self):
        return 'int'


class LongType(IntegralType):
    """Long data type, i.e. a signed 64-bit integer.

    If the values are beyond the range of [-9223372036854775808, 9223372036854775807],
    please use :class:`DecimalType`.
    """

    def simpleString(self):
        return 'bigint'


class ShortType(IntegralType):
    """Short data type, i.e. a signed 16-bit integer.
    """

    def simpleString(self):
        return 'smallint'


class ArrayType(DataType):
    """Array data type.

    :param elementType: :class:`DataType` of each element in the array.
    :param containsNull: boolean, whether the array can contain null (None) values.
    """

    def __init__(self, elementType, containsNull=True):
        """
        >>> ArrayType(StringType()) == ArrayType(StringType(), True)
        True
        >>> ArrayType(StringType(), False) == ArrayType(StringType())
        False
        """
        assert isinstance(elementType, DataType), \
            "elementType %s should be an instance of %s" % (elementType, DataType)
        self.elementType = elementType
        self.containsNull = containsNull

    def simpleString(self):
        return 'array<%s>' % self.elementType.simpleString()

    def __repr__(self):
        return "ArrayType(%s,%s)" % (self.elementType,
                                     str(self.containsNull).lower())

    def jsonValue(self):
        return {"type": self.typeName(),
                "elementType": self.elementType.jsonValue(),
                "containsNull": self.containsNull}

    # noinspection PyShadowingNames
    @classmethod
    def fromJson(cls, json):
        return ArrayType(_parse_datatype_json_value(json["elementType"]),
                         json["containsNull"])

    def needConversion(self):
        return self.elementType.needConversion()

    def toInternal(self, obj):
        if not self.needConversion():
            return obj
        return obj and [self.elementType.toInternal(v) for v in obj]

    def fromInternal(self, obj):
        if not self.needConversion():
            return obj
        return obj and [self.elementType.fromInternal(v) for v in obj]


class MapType(DataType):
    """Map data type.

    :param keyType: :class:`DataType` of the keys in the map.
    :param valueType: :class:`DataType` of the values in the map.
    :param valueContainsNull: indicates whether values can contain null (None) values.

    Keys in a map data type are not allowed to be null (None).
    """

    def __init__(self, keyType, valueType, valueContainsNull=True):
        """
        >>> (MapType(StringType(), IntegerType())
        ...        == MapType(StringType(), IntegerType(), True))
        True
        >>> (MapType(StringType(), IntegerType(), False)
        ...        == MapType(StringType(), FloatType()))
        False
        """
        assert isinstance(keyType, DataType), \
            "keyType %s should be an instance of %s" % (keyType, DataType)
        assert isinstance(valueType, DataType), \
            "valueType %s should be an instance of %s" % (valueType, DataType)
        self.keyType = keyType
        self.valueType = valueType
        self.valueContainsNull = valueContainsNull

    def simpleString(self):
        return 'map<%s,%s>' % (self.keyType.simpleString(), self.valueType.simpleString())

    def __repr__(self):
        return "MapType(%s,%s,%s)" % (self.keyType, self.valueType,
                                      str(self.valueContainsNull).lower())

    def jsonValue(self):
        return {"type": self.typeName(),
                "keyType": self.keyType.jsonValue(),
                "valueType": self.valueType.jsonValue(),
                "valueContainsNull": self.valueContainsNull}

    # noinspection PyShadowingNames
    @classmethod
    def fromJson(cls, json):
        return MapType(_parse_datatype_json_value(json["keyType"]),
                       _parse_datatype_json_value(json["valueType"]),
                       json["valueContainsNull"])

    def needConversion(self):
        return self.keyType.needConversion() or self.valueType.needConversion()

    def toInternal(self, obj):
        if not self.needConversion():
            return obj
        return obj and dict((self.keyType.toInternal(k), self.valueType.toInternal(v))
                            for k, v in obj.items())

    def fromInternal(self, obj):
        if not self.needConversion():
            return obj
        return obj and dict((self.keyType.fromInternal(k), self.valueType.fromInternal(v))
                            for k, v in obj.items())


class StructField(DataType):
    """A field in :class:`StructType`.

    :param name: string, name of the field.
    :param dataType: :class:`DataType` of the field.
    :param nullable: boolean, whether the field can be null (None) or not.
    :param metadata: a dict from string to simple type that can be toInternald to JSON automatically
    """

    def __init__(self, name, dataType, nullable=True, metadata=None):
        """
        >>> (StructField("f1", StringType(), True)
        ...      == StructField("f1", StringType(), True))
        True
        >>> (StructField("f1", StringType(), True)
        ...      == StructField("f2", StringType(), True))
        False
        """
        assert isinstance(dataType, DataType), \
            "dataType %s should be an instance of %s (got %s)" % (dataType, DataType, type(self))
        assert isinstance(name, str), "field name %s should be string" % name
        self.name = name
        self.dataType = dataType
        self.nullable = nullable
        self.metadata = dict(metadata or {})

    def simpleString(self):
        return '%s:%s' % (self.name, self.dataType.simpleString())

    def __repr__(self):
        return "StructField(%s,%s,%s)" % (self.name, self.dataType,
                                          str(self.nullable).lower())

    def jsonValue(self):
        return {"name": self.name,
                "type": self.dataType.jsonValue(),
                "nullable": self.nullable,
                "metadata": self.metadata}

    # noinspection PyShadowingNames
    @classmethod
    def fromJson(cls, json):
        return StructField(json["name"],
                           _parse_datatype_json_value(json["type"]),
                           json["nullable"],
                           json["metadata"])

    def needConversion(self):
        return self.dataType.needConversion()

    def toInternal(self, obj):
        return self.dataType.toInternal(obj)

    def fromInternal(self, obj):
        return self.dataType.fromInternal(obj)

    def typeName(self):
        raise TypeError(
            "StructField does not have typeName. "
            "Use typeName on its type explicitly instead.")


class StructType(DataType):
    """Struct type, consisting of a list of :class:`StructField`.

    This is the data type representing a :class:`Row`.

    Iterating a :class:`StructType` will iterate its :class:`StructField`\\s.
    A contained :class:`StructField` can be accessed by name or position.

    >>> struct1 = StructType([StructField("f1", StringType(), True)])
    >>> struct1["f1"]
    StructField(f1,StringType,true)
    >>> struct1[0]
    StructField(f1,StringType,true)
    """

    def __init__(self, fields=None):
        """
        >>> struct1 = StructType([StructField("f1", StringType(), True)])
        >>> struct2 = StructType([StructField("f1", StringType(), True)])
        >>> struct1 == struct2
        True
        >>> struct1 = StructType([StructField("f1", StringType(), True)])
        >>> struct2 = StructType([StructField("f1", StringType(), True),
        ...     StructField("f2", IntegerType(), False)])
        >>> struct1 == struct2
        False
        """
        if not fields:
            self.fields = []
            self.names = []
        else:
            self.fields = list(fields)
            self.names = [f.name for f in fields]
            assert all(isinstance(f, StructField) for f in fields), \
                "fields should be a list of StructField"
        # Precalculated list of fields that need conversion with fromInternal/toInternal functions
        self._needConversion = [f.needConversion() for f in self]
        self._needSerializeAnyField = any(self._needConversion)

    def add(self, field, data_type=None, nullable=True, metadata=None):
        """
        Construct a StructType by adding new elements to it to define the schema. The method accepts
        either:

            a) A single parameter which is a StructField object.
            b) Between 2 and 4 parameters as (name, data_type, nullable (optional),
               metadata(optional). The data_type parameter may be either a String or a
               DataType object.

        >>> struct1 = StructType().add("f1", StringType(), True).add("f2", StringType(), True, None)
        >>> struct2 = StructType([StructField("f1", StringType(), True), \\
        ...     StructField("f2", StringType(), True, None)])
        >>> struct1 == struct2
        True
        >>> struct1 = StructType().add(StructField("f1", StringType(), True))
        >>> struct2 = StructType([StructField("f1", StringType(), True)])
        >>> struct1 == struct2
        True
        >>> struct1 = StructType().add("f1", "string", True)
        >>> struct2 = StructType([StructField("f1", StringType(), True)])
        >>> struct1 == struct2
        True

        :param field: Either the name of the field or a StructField object
        :param data_type: If present, the DataType of the StructField to create
        :param nullable: Whether the field to add should be nullable (default True)
        :param metadata: Any additional metadata (default None)
        :return: a new updated StructType
        """
        if isinstance(field, StructField):
            self.fields.append(field)
            self.names.append(field.name)
        else:
            if isinstance(field, str) and data_type is None:
                raise ValueError("Must specify DataType if passing name of struct_field to create.")

            if isinstance(data_type, str):
                data_type_f = _parse_datatype_json_value(data_type)
            else:
                data_type_f = data_type
            self.fields.append(StructField(field, data_type_f, nullable, metadata))
            self.names.append(field)
        # Precalculated list of fields that need conversion with fromInternal/toInternal functions
        self._needConversion = [f.needConversion() for f in self]
        self._needSerializeAnyField = any(self._needConversion)
        return self

    def __iter__(self):
        """Iterate the fields"""
        return iter(self.fields)

    def __len__(self):
        """Return the number of fields."""
        return len(self.fields)

    def __getitem__(self, key: Union[str, int, slice]):
        """Access fields by name or slice."""
        if isinstance(key, str):
            for field in self:
                if field.name == key:
                    return field
            raise KeyError(f'No StructField named {key}')
        if isinstance(key, int):
            try:
                return self.fields[key]
            except IndexError as e:
                raise IndexError('StructType index out of range') from e
        if isinstance(key, slice):
            return StructType(self.fields[key])
        raise TypeError('StructType keys should be strings, integers or slices')

    def simpleString(self):
        return 'struct<%s>' % (','.join(f.simpleString() for f in self))

    def treeString(self):
        """
        >>> schema = StructType.fromDDL('some_str: string, some_int: integer, some_date: date')
        >>> print(schema.treeString())
         |-- some_str: string (nullable = true)
         |-- some_int: integer (nullable = true)
         |-- some_date: date (nullable = true)

        >>> schema = StructType.fromDDL('some_str: string, arr: array<string>')
        >>> print(schema.treeString())
         |-- some_str: string (nullable = true)
         |-- arr: array (nullable = true)
         |    |-- element: string (containsNull = true)

        >>> schema = StructType.fromDDL('some_str: string, arr: array<array<string>>')
        >>> print(schema.treeString())
         |-- some_str: string (nullable = true)
         |-- arr: array (nullable = true)
         |    |-- element: array (containsNull = true)
         |    |    |-- element: string (containsNull = true)

        :return: str with the schema inside.
        """
        txt = []
        indent = 0

        def _dump(name: str, type_: DataType, nullable: bool, nullable_name: str = 'nullable') -> str:
            pre = ' |   ' * indent
            return f'{pre} |-- {name}: {type_.typeName()} ({nullable_name} = {"true" if nullable else "false"})'

        def _dump_array(data_type):
            if not isinstance(data_type, ArrayType):
                return []

            nonlocal indent

            txt = []

            indent += 1

            txt.append(
                _dump(
                    'element',
                    data_type.elementType,
                    nullable=data_type.containsNull,
                    nullable_name='containsNull'
                )
            )

            txt.extend(_dump_array(data_type.elementType))

            indent -= 1

            return txt

        for field in self.fields:
            data_type = field.dataType
            txt.append(_dump(field.name, data_type, field.nullable))
            txt.extend(_dump_array(data_type))

        return '\n'.join(txt)

    def __repr__(self):
        return ("StructType(List(%s))" %
                ",".join(str(field) for field in self))

    def jsonValue(self):
        return {"type": self.typeName(),
                "fields": [f.jsonValue() for f in self]}

    # noinspection PyShadowingNames
    @classmethod
    def fromJson(cls, json):
        return StructType([StructField.fromJson(f) for f in json["fields"]])

    def fieldNames(self):
        """
        Returns all field names in a list.

        >>> struct = StructType([StructField("f1", StringType(), True)])
        >>> struct.fieldNames()
        ['f1']
        """
        return list(self.names)

    def needConversion(self):
        # We need convert Row()/namedtuple into tuple()
        return True

    def toInternal(self, obj):
        if obj is None:
            return None

        if self._needSerializeAnyField:
            return self.to_serialized_internal(obj)

        if isinstance(obj, dict):
            return tuple(obj.get(n) for n in self.names)
        if isinstance(obj, Row):
            return obj
        if isinstance(obj, (list, tuple)):
            return tuple(obj)
        if hasattr(obj, "__dict__"):
            d = obj.__dict__
            return tuple(d.get(n) for n in self.names)
        raise ValueError("Unexpected tuple %r with StructType" % obj)

    def to_serialized_internal(self, obj):
        # Only calling toInternal function for fields that need conversion
        if isinstance(obj, dict):
            return tuple(f.toInternal(obj.get(n)) if c else obj.get(n)
                         for n, f, c in zip(self.names, self.fields, self._needConversion))
        if isinstance(obj, Row):
            return create_row(
                obj.__fields__,
                (f.toInternal(val) for f, val, c in zip(self.fields, obj, self._needConversion))
            )
        if isinstance(obj, (tuple, list)):
            return tuple(f.toInternal(v) if c else v
                         for f, v, c in zip(self.fields, obj, self._needConversion))
        if hasattr(obj, "__dict__"):
            d = obj.__dict__
            return tuple(f.toInternal(d.get(n)) if c else d.get(n)
                         for n, f, c in zip(self.names, self.fields, self._needConversion))
        raise ValueError("Unexpected tuple %r with StructType" % obj)

    def fromInternal(self, obj):
        if obj is None:
            return None
        if isinstance(obj, Row):
            # it's already converted by pickler
            return obj
        if self._needSerializeAnyField:
            # Only calling fromInternal function for fields that need conversion
            values = [f.fromInternal(v) if c else v
                      for f, v, c in zip(self.fields, obj, self._needConversion)]
        else:
            values = obj
        return create_row(self.names, values)

    @classmethod
    def fromDDL(cls, string):
        # pylint: disable=import-outside-toplevel, cyclic-import
        from ._ast.ast_to_python import parse_schema
        return parse_schema(string)


class UserDefinedType(DataType):
    """User-defined type (UDT).

    .. note:: WARN: Spark Internal Use Only
    """

    @classmethod
    def typeName(cls):
        return cls.__name__.lower()

    @classmethod
    def sqlType(cls):
        """
        Underlying SQL storage type for this UDT.
        """
        raise NotImplementedError("UDT must implement sqlType().")

    @classmethod
    def module(cls):
        """
        The Python module of the UDT.
        """
        raise NotImplementedError("UDT must implement module().")

    @classmethod
    def scalaUDT(cls):
        """
        The class name of the paired Scala UDT (could be '', if there
        is no corresponding one).
        """
        return ''

    def needConversion(self):
        return True

    @classmethod
    def _cachedSqlType(cls):
        """
        Cache the sqlType() into class, because it's heavy used in `toInternal`.
        """
        if not hasattr(cls, "_cached_sql_type"):
            cls._cached_sql_type = cls.sqlType()
        return cls._cached_sql_type

    def toInternal(self, obj):
        if obj is not None:
            return self._cachedSqlType().toInternal(self.serialize(obj))
        return None

    def fromInternal(self, obj):
        v = self._cachedSqlType().fromInternal(obj)
        if v is not None:
            return self.deserialize(v)
        return None

    def serialize(self, obj):
        """
        Converts the a user-type object into a SQL datum.
        """
        raise NotImplementedError("UDT must implement toInternal().")

    def deserialize(self, datum):
        """
        Converts a SQL datum into a user-type object.
        """
        raise NotImplementedError("UDT must implement fromInternal().")

    def simpleString(self):
        return 'udt'

    def json(self):
        return _json.dumps(self.jsonValue(), separators=(',', ':'), sort_keys=True)

    def jsonValue(self):
        if self.scalaUDT():
            assert self.module() != '__main__', 'UDT in __main__ cannot work with ScalaUDT'
            schema = {
                "type": "udt",
                "class": self.scalaUDT(),
                "pyClass": "%s.%s" % (self.module(), type(self).__name__),
                "sqlType": self.sqlType().jsonValue()
            }
        else:
            raise NotImplementedError("gelanis does not implement jsonValue() for UDT")
            # ser = CloudPickleSerializer()
            # b = ser.dumps(type(self))
            # schema = {
            #     "type": "udt",
            #     "pyClass": "%s.%s" % (self.module(), type(self).__name__),
            #     "serializedClass": base64.b64encode(b).decode('utf8'),
            #     "sqlType": self.sqlType().jsonValue()
            # }
        return schema

    # noinspection PyShadowingNames
    @classmethod
    def fromJson(cls, json):
        pyUDT = str(json["pyClass"])  # convert unicode to str
        split = pyUDT.rfind(".")
        pyModule = pyUDT[:split]
        pyClass = pyUDT[split + 1:]
        m = __import__(pyModule, globals(), locals(), [pyClass])
        if not hasattr(m, pyClass):
            raise NotImplementedError("gelanis does not implement fromJson() for UDT")
            # s = base64.b64decode(json['serializedClass'].encode('utf-8'))
            # UDT = CloudPickleSerializer().loads(s)
        UDT = getattr(m, pyClass)
        return UDT()

    def __eq__(self, other):
        return isinstance(self, type(other))


_atomic_types = [StringType, BinaryType, BooleanType, DecimalType, FloatType, DoubleType,
                 ByteType, ShortType, IntegerType, LongType, DateType, TimestampType, NullType]
_all_atomic_types = dict((t.typeName(), t) for t in _atomic_types)
_all_complex_types = dict((v.typeName(), v)
                          for v in [ArrayType, MapType, StructType])

_FIXED_DECIMAL = re.compile(r"decimal\(\s*(\d+)\s*,\s*(-?\d+)\s*\)")


def _parse_datatype_string(s):
    """
    Parses the given data type string to a :class:`DataType`. The data type string format equals
    to :class:`DataType.simpleString`, except that top level struct type can omit
    the ``struct<>`` and atomic types use ``typeName()`` as their format, e.g. use ``byte`` instead
    of ``tinyint`` for :class:`ByteType`. We can also use ``int`` as a short name
    for :class:`IntegerType`. Since Spark 2.3, this also supports a schema in a DDL-formatted
    string and case-insensitive strings.
    """
    raise NotImplementedError("_parse_datatype_string is not yet supported by gelanis")
    # pylint: disable=W0511
    # todo: implement in pure Python the code below
    # NB: it probably requires to use antl4r

    # sc = Context._active_spark_context
    #
    # def from_ddl_schema(type_str):
    #     return _parse_datatype_json_string(
    #         sc._jvm.org.apache.spark.sql.types.StructType.fromDDL(type_str).json())
    #
    # def from_ddl_datatype(type_str):
    #     return _parse_datatype_json_string(
    #         sc._jvm.org.apache.spark.sql.api.python.PythonSQLUtils.parseDataType(type_str).json())
    #
    # try:
    #     # DDL format, "fieldname datatype, fieldname datatype".
    #     return from_ddl_schema(s)
    # except Exception as e:
    #     try:
    #         # For backwards compatibility, "integer", "struct<fieldname: datatype>" and etc.
    #         return from_ddl_datatype(s)
    #     except:
    #         try:
    #             # For backwards compatibility, "fieldname: datatype, fieldname: datatype" case.
    #             return from_ddl_datatype("struct<%s>" % s.strip())
    #         except:
    #             raise e


def _parse_datatype_json_string(json_string):
    """Parses the given data type JSON string."""
    return _parse_datatype_json_value(_json.loads(json_string))


def _parse_datatype_json_value(json_value):
    if not isinstance(json_value, dict):
        if json_value in _all_atomic_types.keys():
            return _all_atomic_types[json_value]()
        if json_value == 'decimal':
            return DecimalType()
        if _FIXED_DECIMAL.match(json_value):
            m = _FIXED_DECIMAL.match(json_value)
            return DecimalType(int(m.group(1)), int(m.group(2)))
        raise ValueError("Could not parse datatype: %s" % json_value)
    tpe = json_value["type"]
    if tpe in _all_complex_types:
        return _all_complex_types[tpe].fromJson(json_value)
    if tpe == 'udt':
        return UserDefinedType.fromJson(json_value)
    raise ValueError("not supported type: %s" % tpe)


# Mapping Python types to Spark SQL DataType
_type_mappings = {
    type(None): NullType,
    bool: BooleanType,
    int: LongType,
    float: DoubleType,
    str: StringType,
    bytearray: BinaryType,
    decimal.Decimal: DecimalType,
    datetime.date: DateType,
    datetime.datetime: TimestampType,
    datetime.time: TimestampType,
    bytes: BinaryType,
}

# Mapping Python array types to Spark SQL DataType
# We should be careful here. The size of these types in python depends on C
# implementation. We need to make sure that this conversion does not lose any
# precision. Also, JVM only support signed types, when converting unsigned types,
# keep in mind that it required 1 more bit when stored as singed types.
#
# Reference for C integer size, see:
# ISO/IEC 9899:201x specification, chapter 5.2.4.2.1 Sizes of integer types <limits.h>.
# Reference for python array typecode, see:
# https://docs.python.org/2/library/array.html
# https://docs.python.org/3.6/library/array.html
# Reference for JVM's supported integral types:
# http://docs.oracle.com/javase/specs/jvms/se8/html/jvms-2.html#jvms-2.3.1

_array_signed_int_typecode_ctype_mappings = {
    'b': ctypes.c_byte,
    'h': ctypes.c_short,
    'i': ctypes.c_int,
    'l': ctypes.c_long,
}

_array_unsigned_int_typecode_ctype_mappings = {
    'B': ctypes.c_ubyte,
    'H': ctypes.c_ushort,
    'I': ctypes.c_uint,
    'L': ctypes.c_ulong
}


# noinspection PyShadowingNames
def _int_size_to_type(size):
    """
    Return the Catalyst datatype from the size of integers.
    """
    if size <= 8:
        return ByteType
    if size <= 16:
        return ShortType
    if size <= 32:
        return IntegerType
    if size <= 64:
        return LongType
    return None


# The list of all supported array typecodes is stored here
_array_type_mappings = {
    # Warning: Actual properties for float and double in C is not specified in C.
    # On almost every system supported by both python and JVM, they are IEEE 754
    # single-precision binary floating-point format and IEEE 754 double-precision
    # binary floating-point format. And we do assume the same thing here for now.
    'f': FloatType,
    'd': DoubleType
}

# compute array typecode mappings for signed integer types
for _typecode in _array_signed_int_typecode_ctype_mappings:
    _size = ctypes.sizeof(_array_signed_int_typecode_ctype_mappings[_typecode]) * 8
    _dt = _int_size_to_type(_size)
    if _dt is not None:
        _array_type_mappings[_typecode] = _dt

# compute array typecode mappings for unsigned integer types
for _typecode in _array_unsigned_int_typecode_ctype_mappings:
    # JVM does not have unsigned types, so use signed types that is at least 1
    # bit larger to store
    _size = ctypes.sizeof(_array_unsigned_int_typecode_ctype_mappings[_typecode]) * 8 + 1
    _dt = _int_size_to_type(_size)
    if _dt is not None:
        _array_type_mappings[_typecode] = _dt

# Type code 'u' in Python's array is deprecated since version 3.3, and will be
# removed in version 4.0. See: https://docs.python.org/3/library/array.html
if sys.version_info[0] < 4:
    _array_type_mappings['u'] = StringType


def _infer_type(obj):
    """Infer the DataType from obj
    """
    if obj is None:
        return NullType()

    if hasattr(obj, '__UDT__'):
        return obj.__UDT__

    dataType = _type_mappings.get(type(obj))
    if dataType is DecimalType:
        # the precision and scale of `obj` may be different from row to row.
        return DecimalType(38, 18)
    if dataType is not None:
        return dataType()

    struct_type = _infer_struct_type(obj)
    if struct_type is not None:
        return struct_type

    try:
        return _infer_schema(obj)
    except TypeError as e:
        raise TypeError("not supported type: %s" % type(obj)) from e


def _infer_struct_type(obj):
    if isinstance(obj, dict):
        for key, value in obj.items():
            if key is not None and value is not None:
                return MapType(_infer_type(key), _infer_type(value), True)
        return MapType(NullType(), NullType(), True)
    if isinstance(obj, list):
        for v in obj:
            if v is not None:
                return ArrayType(_infer_type(obj[0]), True)
        return ArrayType(NullType(), True)
    if isinstance(obj, array):
        if obj.typecode in _array_type_mappings:
            return ArrayType(_array_type_mappings[obj.typecode](), False)
        raise TypeError("not supported type: array(%s)" % obj.typecode)
    return None


def _infer_schema(row, names=None):
    """Infer the schema from dict/namedtuple/object"""
    if isinstance(row, dict):
        items = sorted(row.items())

    elif isinstance(row, (tuple, list)):
        if hasattr(row, "__fields__"):  # Row
            items = zip(row.__fields__, tuple(row))
        elif hasattr(row, "_fields"):  # namedtuple
            # noinspection PyProtectedMember
            items = zip(row._fields, tuple(row))
        else:
            if names is None:
                names = ['_%d' % i for i in range(1, len(row) + 1)]
            elif len(names) < len(row):
                names.extend('_%d' % i for i in range(len(names) + 1, len(row) + 1))
            items = zip(names, row)

    elif hasattr(row, "__dict__"):  # object
        items = sorted(row.__dict__.items())

    else:
        raise TypeError("Can not infer schema for type: %s" % type(row))

    fields = [StructField(k, _infer_type(v), True) for k, v in items]
    return StructType(fields)


# noinspection PyShadowingNames
def _has_nulltype(dt):
    """ Return whether there is NullType in `dt` or not """
    if isinstance(dt, StructType):
        return any(_has_nulltype(f.dataType) for f in dt.fields)
    if isinstance(dt, ArrayType):
        return _has_nulltype(dt.elementType)
    if isinstance(dt, MapType):
        return _has_nulltype(dt.keyType) or _has_nulltype(dt.valueType)
    return isinstance(dt, NullType)


def _get_null_fields(field, prefix=""):
    """ Return all field names which have a NullType in `field`"""
    if isinstance(field, StructType):
        return list(itertools.chain(*(
            _get_null_fields(f) for f in field.fields
        )))

    if prefix:
        prefixed_field_name = f"{prefix}.{field.name}"
    else:
        prefixed_field_name = field.name

    if isinstance(field, ArrayType):
        return _get_null_fields(field.elementType, prefix=prefixed_field_name + ".elements")

    if isinstance(field, MapType):
        return [
            _get_null_fields(field.keyType, prefix=prefixed_field_name + ".keys")
            + _get_null_fields(field.valueType, prefix=prefixed_field_name + ".values")
        ]

    return [prefixed_field_name] if isinstance(field.dataType, NullType) else []


def _merge_type(a, b, name=None):
    if name is None:
        new_msg = lambda msg: msg
        new_name = lambda n: "field %s" % n
    else:
        new_msg = lambda msg: "%s: %s" % (name, msg)
        new_name = lambda n: "field %s in %s" % (n, name)

    if isinstance(a, NullType):
        return b
    if isinstance(b, NullType):
        return a
    if type(a) is not type(b):
        # pylint: disable=W0511
        # TODO: type cast (such as int -> long)
        raise TypeError(new_msg("Can not merge type %s and %s" % (type(a), type(b))))

    # same type
    if isinstance(a, StructType):
        nfs = dict((f.name, f.dataType) for f in b.fields)
        fields = [StructField(f.name, _merge_type(f.dataType, nfs.get(f.name, NullType()),
                                                  name=new_name(f.name)))
                  for f in a.fields]
        names = {f.name for f in fields}
        for n in nfs:
            if n not in names:
                fields.append(StructField(n, nfs[n]))
        return StructType(fields)

    if isinstance(a, ArrayType):
        return ArrayType(_merge_type(a.elementType, b.elementType,
                                     name='element in array %s' % name), True)

    if isinstance(a, MapType):
        return MapType(_merge_type(a.keyType, b.keyType, name='key of map %s' % name),
                       _merge_type(a.valueType, b.valueType, name='value of map %s' % name),
                       True)
    return a


def _need_converter(dataType):
    if isinstance(dataType, StructType):
        return True
    if isinstance(dataType, ArrayType):
        return _need_converter(dataType.elementType)
    if isinstance(dataType, MapType):
        return _need_converter(dataType.keyType) or _need_converter(dataType.valueType)
    if isinstance(dataType, NullType):
        return True
    return False


def _create_converter(dataType):
    """Create a converter to drop the names of fields in obj
    :type dataType: Union[DataType, StructField]
    """
    if not _need_converter(dataType):
        return lambda x: x

    if isinstance(dataType, ArrayType):
        conv = _create_converter(dataType.elementType)
        return lambda row: [conv(v) for v in row]

    if isinstance(dataType, MapType):
        kconv = _create_converter(dataType.keyType)
        vconv = _create_converter(dataType.valueType)
        return lambda row: dict((kconv(k), vconv(v)) for k, v in row.items())

    if isinstance(dataType, NullType):
        return lambda x: None

    if not isinstance(dataType, StructType):
        return lambda x: x

    # dataType must be StructType
    names = [f.name for f in dataType.fields]
    converters = [_create_converter(f.dataType) for f in dataType.fields]
    convert_fields = any(_need_converter(f.dataType) for f in dataType.fields)

    def convert_struct(obj):
        if obj is None:
            return None

        if isinstance(obj, Row):
            if convert_fields:
                return create_row(
                    obj.__fields__,
                    [converter(value) for converter, value in zip(converters, obj)]
                )
            return obj

        if isinstance(obj, (tuple, list)):
            if convert_fields:
                return tuple(c(v) for v, c in zip(obj, converters))
            return tuple(obj)

        return convert_dict(obj)

    def convert_dict(obj):
        if isinstance(obj, dict):
            d = obj
        elif hasattr(obj, "__dict__"):  # object
            d = obj.__dict__
        else:
            raise TypeError("Unexpected obj type: %s" % type(obj))

        if convert_fields:
            return tuple(convert(d.get(name)) for name, convert in zip(names, converters))

        return tuple(d.get(name) for name in names)

    return convert_struct


_acceptable_types = {
    BooleanType: (bool,),
    ByteType: (int,),
    ShortType: (int,),
    IntegerType: (int,),
    LongType: (int,),
    FloatType: (float,),
    DoubleType: (float,),
    DecimalType: (decimal.Decimal,),
    StringType: (str,),
    BinaryType: (bytearray, bytes),
    DateType: (datetime.date, datetime.datetime,),
    TimestampType: (datetime.datetime,),
    ArrayType: (list, tuple, array,),
    MapType: (dict,),
    StructType: (tuple, list, dict,),
}


# pylint: disable=too-many-locals, too-many-branches, too-many-statements
def _make_type_verifier(dataType, nullable=True, name=None):
    """
    Make a verifier that checks the type of obj against dataType and raises a TypeError if they do
    not match.

    This verifier also checks the value of obj against datatype and raises a ValueError if it's not
    within the allowed range, e.g. using 128 as ByteType will overflow. Note that, Python float is
    not checked, so it will become infinity when cast to Java float if it overflows.

    >>> _make_type_verifier(StructType([]))(None)
    >>> _make_type_verifier(StringType())("")
    >>> _make_type_verifier(LongType())(0)
    >>> _make_type_verifier(ArrayType(ShortType()))(list(range(3)))
    >>> _make_type_verifier(ArrayType(StringType()))(set()) # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    TypeError:...
    >>> _make_type_verifier(MapType(StringType(), IntegerType()))({})
    >>> _make_type_verifier(StructType([]))(())
    >>> _make_type_verifier(StructType([]))([])
    >>> _make_type_verifier(StructType([]))([1]) # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    ValueError:...
    >>> # Check if numeric values are within the allowed range.
    >>> _make_type_verifier(ByteType())(12)
    >>> _make_type_verifier(ByteType())(1234) # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    ValueError:...
    >>> _make_type_verifier(ByteType(), False)(None) # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    ValueError:...
    >>> _make_type_verifier(
    ...  ArrayType(ShortType(), False)
    ... )([1, None]) # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    ValueError:...
    >>> _make_type_verifier(
    ...  MapType(StringType(), IntegerType())
    ... )({None: 1}) # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    ValueError:...
    >>> schema = StructType().add("a", IntegerType()).add("b", StringType(), False)
    >>> _make_type_verifier(schema)((1, None)) # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    ValueError:...
    """

    if name is None:
        new_msg = lambda msg: msg
        new_name = lambda n: "field %s" % n
    else:
        new_msg = lambda msg: "%s: %s" % (name, msg)
        new_name = lambda n: "field %s in %s" % (n, name)

    def verify_nullability(obj):
        if obj is None:
            if nullable:
                return True

            raise ValueError(new_msg("This field is not nullable, but got None"))

        return False

    _type = type(dataType)

    def assert_acceptable_types(obj):
        assert _type in _acceptable_types, \
            new_msg("unknown datatype: %s for object %r" % (dataType, obj))

    def verify_acceptable_types(obj):
        # subclass of them can not be fromInternal in JVM
        if type(obj) not in _acceptable_types[_type]:
            raise TypeError(new_msg("%s can not accept object %r in type %s"
                                    % (dataType, obj, type(obj))))

    if isinstance(dataType, StringType):
        # StringType can work with any types
        verify_value = lambda _: _

    elif isinstance(dataType, UserDefinedType):
        verifier = _make_type_verifier(dataType.sqlType(), name=name)

        def verify_udf(obj):
            if not (hasattr(obj, '__UDT__') and obj.__UDT__ == dataType):
                raise ValueError(new_msg("%r is not an instance of type %r" % (obj, dataType)))
            verifier(dataType.toInternal(obj))

        verify_value = verify_udf

    elif isinstance(dataType, ByteType):
        def verify_byte(obj):
            assert_acceptable_types(obj)
            verify_acceptable_types(obj)
            if obj < -128 or obj > 127:
                raise ValueError(new_msg("object of ByteType out of range, got: %s" % obj))

        verify_value = verify_byte

    elif isinstance(dataType, ShortType):
        def verify_short(obj):
            assert_acceptable_types(obj)
            verify_acceptable_types(obj)
            if obj < -32768 or obj > 32767:
                raise ValueError(new_msg("object of ShortType out of range, got: %s" % obj))

        verify_value = verify_short

    elif isinstance(dataType, IntegerType):
        def verify_integer(obj):
            assert_acceptable_types(obj)
            verify_acceptable_types(obj)
            if obj < -2147483648 or obj > 2147483647:
                raise ValueError(
                    new_msg("object of IntegerType out of range, got: %s" % obj))

        verify_value = verify_integer

    elif isinstance(dataType, LongType):
        def verify_long(obj):
            assert_acceptable_types(obj)
            verify_acceptable_types(obj)
            if obj < -9223372036854775808 or obj > 9223372036854775807:
                raise ValueError(
                    new_msg("object of LongType out of range, got: %s" % obj))

        verify_value = verify_long

    elif isinstance(dataType, ArrayType):
        element_verifier = _make_type_verifier(
            dataType.elementType, dataType.containsNull, name="element in array %s" % name)

        def verify_array(obj):
            assert_acceptable_types(obj)
            verify_acceptable_types(obj)
            for i in obj:
                element_verifier(i)

        verify_value = verify_array

    elif isinstance(dataType, MapType):
        key_verifier = _make_type_verifier(dataType.keyType, False, name="key of map %s" % name)
        value_verifier = _make_type_verifier(
            dataType.valueType, dataType.valueContainsNull, name="value of map %s" % name)

        def verify_map(obj):
            assert_acceptable_types(obj)
            verify_acceptable_types(obj)
            for k, v in obj.items():
                key_verifier(k)
                value_verifier(v)

        verify_value = verify_map

    elif isinstance(dataType, StructType):
        verifiers = []
        for f in dataType.fields:
            verifier = _make_type_verifier(f.dataType, f.nullable, name=new_name(f.name))
            verifiers.append((f.name, verifier))

        def verify_struct(obj):
            assert_acceptable_types(obj)

            if isinstance(obj, dict):
                for f, verifier in verifiers:
                    verifier(obj.get(f))
            elif isinstance(obj, (tuple, list)):
                if len(obj) != len(verifiers):
                    raise ValueError(
                        new_msg("Length of object (%d) does not match with "
                                "length of fields (%d)" % (len(obj), len(verifiers))))
                for v, (_, verifier) in zip(obj, verifiers):
                    verifier(v)
            elif hasattr(obj, "__dict__"):
                d = obj.__dict__
                for f, verifier in verifiers:
                    verifier(d.get(f))
            else:
                raise TypeError(new_msg("StructType can not accept object %r in type %s"
                                        % (obj, type(obj))))
        verify_value = verify_struct

    else:
        def verify_default(obj):
            assert_acceptable_types(obj)
            verify_acceptable_types(obj)

        verify_value = verify_default

    def verify(obj):
        if not verify_nullability(obj):
            verify_value(obj)

    return verify


# As defined with visitPrimitiveDataType in spark master sql/catalyst/src/main/scala/
# org/apache/spark/sql/catalyst/parser/AstBuilder.scala
STRING_TO_TYPE = dict(
    boolean=BooleanType(),
    tinyint=ByteType(),
    byte=ByteType(),
    smallint=ShortType(),
    short=ShortType(),
    int=IntegerType(),
    integer=IntegerType(),
    bigint=LongType(),
    long=LongType(),
    float=FloatType(),
    real=FloatType(),
    double=DoubleType(),
    date=DateType(),
    timestamp=TimestampType(),
    string=StringType(),
    varchar=StringType(),
    char=StringType(),
    binary=BinaryType(),
    decimal=DecimalType(),
    dec=DecimalType(),
    numeric=DecimalType(),
    void=NullType(),
)


def parsed_string_to_type(data_type, arguments):
    data_type = data_type.lower()
    if not arguments and data_type in STRING_TO_TYPE:
        return STRING_TO_TYPE[data_type]

    if data_type in ['varchar', 'char']:
        return STRING_TO_TYPE[data_type]

    if data_type in ["decimal", 'dec', 'numeric']:
        if len(arguments) == 2:
            precision, scale = arguments
        elif len(arguments) == 1:
            precision, scale = arguments[0], 0
        else:
            raise ParseException("Unrecognized decimal parameters: {0}".format(arguments))
        return DecimalType(precision=int(precision), scale=int(scale))

    if data_type == "array" and len(arguments) == 1:
        return ArrayType(arguments[0])

    if data_type == "map" and len(arguments) == 2:
        return MapType(arguments[0], arguments[1])

    if data_type == "struct":
        return StructType([
            StructField(name, data_type, metadata=ChainMap(*metadata))
            for name, data_type, *metadata in (arguments[0] if arguments else [])
        ])

    raise ParseException(
        "Unable to parse data type {0}{1}".format(data_type, arguments if arguments else "")
    )


# Internal type hierarchy:
# Lower order types are converted to first order type when performing operation
# across multiple types (e.g. datetime.date(2019,1,1) == "2019-01-01")
INTERNAL_TYPE_ORDER = [
    float,
    int,
    str,
    bool,
    datetime.datetime,
    datetime.date,
]

PYTHON_TO_SPARK_TYPE = {
    type(None): NullType(),
    bool: BooleanType(),
    int: IntegerType(),
    float: DoubleType(),
    str: StringType(),
    bytearray: BinaryType,
    bytes: BinaryType,
    decimal.Decimal: DecimalType(),
    datetime.date: DateType(),
    datetime.datetime: TimestampType(),
    datetime.time: TimestampType(),
}


def python_to_spark_type(python_type):
    """
    :type python_type: type
    :rtype DataType
    """
    # pylint: disable=W0511
    # todo: support Decimal
    if python_type in PYTHON_TO_SPARK_TYPE:
        return PYTHON_TO_SPARK_TYPE[python_type]
    raise NotImplementedError(
        f"Pysparkling does not currently support type {python_type} for the requested operation"
    )



class Options(dict):
    """
    A case insensitive dict, which can be initialized from multiple dicts
    and whose values can be access through attr syntax

    It also stores "false" and "true" strings as Boolean

    e.g.:

    >>> default_options = dict(sep=",", samplingRatio=None)
    >>> requested_options = dict(Sep="|")
    >>> o=Options({"format": "json", "lineSep": ","}, Format="csv")
    >>> o.format, o.linesep
    ('csv', ',')
    >>> o.UndefinedSetting
    Traceback (most recent call last):
    ...
    KeyError: 'undefinedsetting'
    """

    def __init__(self, *args, **kwargs):
        d = {
            key.lower(): value
            for arg in args
            if arg is not None
            for key, value in arg.items()
        }
        d.update({
            key.lower(): value
            for key, value in kwargs.items()
        })
        super().__init__(d)

    def setdefault(self, k, default=None):
        return super().setdefault(k.lower(), default)

    @staticmethod
    def fromkeys(seq, value=None):
        return Options({k.lower(): value for k in seq})

    def __getitem__(self, k):
        return super().__getitem__(k.lower())

    def __setitem__(self, k, v):
        if isinstance(v, str) and v.lower() in ("true", "false"):
            v = (v.lower() == "true")
        super().__setitem__(k.lower(), v)

    def __delitem__(self, k):
        super().__delitem__(k.lower())

    def get(self, k, *args, **kwargs):
        return super().get(k.lower(), *args, **kwargs)

    def __contains__(self, o):
        if not isinstance(o, str):
            return False
        return super().__contains__(o.lower())

    def __getattr__(self, item):
        if not item.startswith("_"):
            return self[item.lower()]
        return getattr(super(), item)



__all__ = ['SparkConf']


class SparkConf:

    """
    Configuration for a Spark application. Used to set various Spark
    parameters as key-value pairs.

    Most of the time, you would create a SparkConf object with
    ``SparkConf()``, which will load values from `spark.*` Java system
    properties as well. In this case, any parameters you set directly on
    the :class:`SparkConf` object take priority over system properties.

    For unit tests, you can also call ``SparkConf(false)`` to skip
    loading external settings and get the same configuration no matter
    what the system properties are.

    All setter methods in this class support chaining. For example,
    you can write ``conf.setMaster("local").setAppName("My app")``.

    Parameters
    ----------
    loadDefaults : bool
        whether to load values from Java system properties (True by default)

    Notes
    -----
    Once a SparkConf object is passed to Spark, it is cloned
    and can no longer be modified by the user.

    Examples
    --------
    >>> from gelanis.conf import SparkConf
    >>> from gelanis.context import SparkContext
    >>> conf = SparkConf()
    >>> conf.setMaster("local").setAppName("My app")  # doctest: +ELLIPSIS
    <gelanis.conf.SparkConf object at ...>
    >>> conf.get("spark.master")
    'local'
    >>> conf.get("spark.app.name")
    'My app'
    >>> sc = SparkContext(conf=conf)
    >>> sc.master
    'local'
    >>> sc.appName
    'My app'
    >>> sc.sparkHome is None
    True

    >>> conf = SparkConf(loadDefaults=False)
    >>> conf.setSparkHome("/path")  # doctest: +ELLIPSIS
    <gelanis.conf.SparkConf object at ...>
    >>> conf.get("spark.home")
    '/path'
    >>> conf.setExecutorEnv("VAR1", "value1")  # doctest: +ELLIPSIS
    <gelanis.conf.SparkConf object at ...>
    >>> conf.setExecutorEnv(pairs = [("VAR3", "value3"), ("VAR4", "value4")])  # doctest: +ELLIPSIS
    <gelanis.conf.SparkConf object at ...>
    >>> conf.get("spark.executorEnv.VAR1")
    'value1'
    >>> print(conf.toDebugString())
    spark.home=/path
    spark.executorEnv.VAR1=value1
    spark.executorEnv.VAR3=value3
    spark.executorEnv.VAR4=value4
    >>> for p in sorted(conf.getAll(), key=lambda p: p[0]):
    ...     print(p)
    ('spark.executorEnv.VAR1', 'value1')
    ('spark.executorEnv.VAR3', 'value3')
    ('spark.executorEnv.VAR4', 'value4')
    ('spark.home', '/path')
    >>> print(conf.toDebugString())
    spark.home=/path
    spark.executorEnv.VAR1=value1
    spark.executorEnv.VAR3=value3
    spark.executorEnv.VAR4=value4
    """

    def __init__(self, loadDefaults=True, _jvm=None, _jconf=None):
        """
        Create a new Spark configuration.
        """
        if _jconf:
            self._jconf = _jconf
        else:
            self._jconf = None
            self._conf = {}

    def set(self, key, value):
        """Set a configuration property."""
        # Try to set self._jconf first if JVM is created, set self._conf if JVM is not created yet.
        if self._jconf is not None:
            self._jconf.set(key, str(value))
        else:
            self._conf[key] = str(value)
        return self

    def setIfMissing(self, key, value):
        """Set a configuration property, if not already set."""
        if self.get(key) is None:
            self.set(key, value)
        return self

    def setMaster(self, value):
        """Set master URL to connect to."""
        self.set("spark.master", value)
        return self

    def setAppName(self, value):
        """Set application name."""
        self.set("spark.app.name", value)
        return self

    def setSparkHome(self, value):
        """Set path where Spark is installed on worker nodes."""
        self.set("spark.home", value)
        return self

    def setExecutorEnv(self, key=None, value=None, pairs=None):
        """Set an environment variable to be passed to executors."""
        if (key is not None and pairs is not None) or (key is None and pairs is None):
            raise Exception("Either pass one key-value pair or a list of pairs")

        if key is not None:
            self.set("spark.executorEnv." + key, value)
        elif pairs is not None:
            for (k, v) in pairs:
                self.set("spark.executorEnv." + k, v)

        return self

    def setAll(self, pairs):
        """
        Set multiple parameters, passed as a list of key-value pairs.

        Parameters
        ----------
        pairs : iterable of tuples
            list of key-value pairs to set
        """
        for (k, v) in pairs:
            self.set(k, v)
        return self

    def get(self, key, defaultValue=None):
        """Get the configured value for some key, or return a default otherwise."""
        if defaultValue is None:   # Py4J doesn't call the right get() if we pass None
            if self._jconf is not None:
                if not self._jconf.contains(key):
                    return None
                return self._jconf.get(key)

            if key not in self._conf:
                return None
            return self._conf[key]

        if self._jconf is not None:
            return self._jconf.get(key, defaultValue)

        return self._conf.get(key, defaultValue)

    def getAll(self):
        """Get all values as a list of key-value pairs."""
        if self._jconf is not None:
            return [(elem._1(), elem._2()) for elem in self._jconf.getAll()]

        return self._conf.items()

    def contains(self, key):
        """Does this configuration contain a given key?"""
        if self._jconf is not None:
            return self._jconf.contains(key)

        return key in self._conf

    def toDebugString(self):
        """
        Returns a printable version of the configuration, as a list of
        key=value pairs, one per line.
        """
        if self._jconf is not None:
            return self._jconf.toDebugString()

        return '\n'.join('%s=%s' % (k, v) for k, v in self._conf.items())



from threading import RLock

from ..__version__ import __version__
from ..context import Context
from ..rdd import RDD
from ._schema_utils import infer_schema_from_list
from .conf import RuntimeConfig
from .dataframe import DataFrame
from .internals import DataFrameInternal
from .pandas.utils import require_minimum_pandas_version
from .readwriter import DataFrameReader
from .types import (
    _create_converter, _has_nulltype, _infer_schema, _make_type_verifier, _merge_type, DataType, StructType
)

__all__ = ['SparkSession']


def _monkey_patch_RDD(sparkSession):
    def toDF(self, schema=None, sampleRatio=None):
        """
        Converts current :class:`RDD` into a :class:`DataFrame`

        This is a shorthand for ``spark.createDataFrame(rdd, schema, sampleRatio)``

        Parameters
        ----------
        schema : :class:`pyspark.sql.types.DataType`, str or list, optional
            a :class:`pyspark.sql.types.DataType` or a datatype string or a list of
            column names, default is None.  The data type string format equals to
            :class:`pyspark.sql.types.DataType.simpleString`, except that top level struct type can
            omit the ``struct<>`` and atomic types use ``typeName()`` as their format, e.g. use
            ``byte`` instead of ``tinyint`` for :class:`pyspark.sql.types.ByteType`.
            We can also use ``int`` as a short name for :class:`pyspark.sql.types.IntegerType`.
        sampleRatio : float, optional
            the sample ratio of rows used for inferring

        Returns
        -------
        :class:`DataFrame`

        Examples
        --------
        >>> rdd.toDF().collect()
        [Row(name='Alice', age=1)]
        """
        return sparkSession.createDataFrame(self, schema, sampleRatio)

    RDD.toDF = toDF


class SparkSession:
    class Builder:
        _lock = RLock()
        _options = {}

        def config(self, key=None, value=None, conf=None):
            """Sets a config option. Options set using this method are automatically propagated to
            both :class:`SparkConf` and :class:`SparkSession`'s own configuration.

            .. versionadded:: 2.0.0

            Parameters
            ----------
            key : str, optional
                a key name string for configuration property
            value : str, optional
                a value for configuration property
            conf : :class:`SparkConf`, optional
                an instance of :class:`SparkConf`

            Examples
            --------
            For an existing SparkConf, use `conf` parameter.

            >>> from gelanis.conf import SparkConf
            >>> SparkSession.builder.config(conf=SparkConf())  # doctest: +ELLIPSIS
            <gelanis.sql.session...

            For a (key, value) pair, you can omit parameter names.

            >>> SparkSession.builder.config("spark.some.config.option", "some-value")  # doctest: +ELLIPSIS
            <gelanis.sql.session...

            """
            with self._lock:
                if conf is None:
                    self._options[key] = str(value)
                else:
                    for (k, v) in conf.getAll():
                        self._options[k] = v
                return self

        def master(self, master):
            """Sets the Spark master URL to connect to, such as "local" to run locally, "local[4]"
            to run locally with 4 cores, or "spark://master:7077" to run on a Spark standalone
            cluster.

            .. versionadded:: 2.0.0

            Parameters
            ----------
            master : str
                a url for spark master
            """
            return self.config("spark.master", master)

        def appName(self, name):
            """Sets a name for the application, which will be shown in the Spark web UI.

            If no application name is set, a randomly generated name will be used.

            .. versionadded:: 2.0.0

            Parameters
            ----------
            name : str
                an application name
            """
            return self.config("spark.app.name", name)

        def getOrCreate(self):
            with self._lock:
                session = SparkSession._instantiatedSession
                if session is None:
                    session = SparkSession(Context())
                return session

    _instantiatedSession = None
    _activeSession = None

    builder = Builder()

    def __init__(self, sparkContext, jsparkSession=None):
        # pylint: disable=import-outside-toplevel, cyclic-import
        from .context import SQLContext

        self._sc = sparkContext
        self._wrapped = SQLContext(self._sc, self)
        SparkSession._instantiatedSession = self
        SparkSession._activeSession = self

        _monkey_patch_RDD(self)

    def newSession(self):
        """
        Returns a new SparkSession as new session, that has separate SQLConf,
        registered temporary views and UDFs, but shared Context and
        table cache.
        """
        return self.__class__(self._sc)

    @classmethod
    def getActiveSession(cls):
        return SparkSession._activeSession

    @property
    def sparkContext(self):
        """Returns the underlying Context."""
        return self._sc

    @property
    def version(self):
        return __version__

    @property
    def conf(self):
        """Runtime configuration interface for Spark.

        This is the interface through which the user can get and set all Spark and Hadoop
        configurations that are relevant to Spark SQL. When getting the value of a config,
        this defaults to the value set in the underlying :class:`Context`, if any.
        """
        if not hasattr(self, "_conf"):
            # Compatibility with Pyspark behavior
            # noinspection PyAttributeOutsideInit
            # pylint: disable=W0201
            self._conf = RuntimeConfig()
        return self._conf

    @property
    def catalog(self):
        """Interface through which the user may create, drop, alter or query underlying
        databases, tables, functions etc.

        :return: :class:`Catalog`
        """
        # from .catalog import Catalog
        # if not hasattr(self, "_catalog"):
        #     # Compatibility with Pyspark behavior
        #     # noinspection PyAttributeOutsideInit
        #     self._catalog = Catalog(self)
        # return self._catalog
        raise NotImplementedError()

    @property
    def udf(self):
        # pylint: disable=W0511
        # todo: Add support of udf registration
        raise NotImplementedError("Pysparkling does not support yet catalog")
        # from .udf import UDFRegistration
        # return UDFRegistration(self)

    def _inferSchema(self, rdd, samplingRatio=None, names=None):
        """
        Infer schema from an RDD of Row or tuple.

        :param rdd: an RDD of Row or tuple
        :param samplingRatio: sampling ratio, or no sampling (default)
        :return: :class:`.types.StructType`
        """
        first = rdd.first()
        if not first:
            raise ValueError("The first row in RDD is empty, "
                             "can not infer schema")
        if isinstance(first, dict):
            raise NotImplementedError(
                "Using RDD of dict to inferSchema is deprecated in Spark "
                "and not implemented in gelanis. "
                "Please use .sql.Row instead"
            )

        if samplingRatio is None:
            schema = _infer_schema(first, names=names)
            if _has_nulltype(schema):
                for row in rdd.take(100)[1:]:
                    schema = _merge_type(schema, _infer_schema(row, names=names))
                    if not _has_nulltype(schema):
                        break
                else:
                    raise ValueError("Some of types cannot be determined by the "
                                     "first 100 rows, please try again with sampling")
        else:
            if samplingRatio < 0.99:
                rdd = rdd.sample(False, float(samplingRatio))
            schema = rdd.map(lambda r: _infer_schema(r, names)).reduce(_merge_type)
        return schema

    def _createFromRDD(self, rdd, schema, samplingRatio):
        """
        Create an RDD for DataFrame from an existing RDD, returns the RDD and schema.
        """
        if schema is None or isinstance(schema, (list, tuple)):
            struct = self._inferSchema(rdd, samplingRatio, names=schema)
            converter = _create_converter(struct)
            rdd = rdd.map(converter)
            if isinstance(schema, (list, tuple)):
                for i, name in enumerate(schema):
                    struct.fields[i].name = name
                    struct.names[i] = name
            schema = struct

        elif not isinstance(schema, StructType):
            raise TypeError("schema should be StructType or list or None, but got: %s" % schema)

        # convert python objects to sql data
        rdd = rdd.map(schema.toInternal)
        return rdd, schema

    def _createFromLocal(self, data, schema):
        """
        Create an RDD for DataFrame from a list or pandas.DataFrame, returns
        the RDD and schema.
        """
        # make sure data could consumed multiple times
        if not isinstance(data, list):
            data = list(data)

        if schema is None or isinstance(schema, (list, tuple)):
            struct = infer_schema_from_list(data, names=schema)
            converter = _create_converter(struct)
            data = map(converter, data)
            if isinstance(schema, (list, tuple)):
                for i, name in enumerate(schema):
                    struct.fields[i].name = name
                    struct.names[i] = name
            schema = struct

        elif not isinstance(schema, StructType):
            raise TypeError("schema should be StructType or list or None, but got: %s" % schema)

        # convert python objects to sql data
        data = [schema.toInternal(row) for row in data]
        return self._sc.parallelize(data), schema

    # noinspection PyMethodMayBeStatic
    def _get_numpy_record_dtype(self, rec):
        # numpy is an optional dependency
        # pylint: disable=import-outside-toplevel
        import numpy as np
        cur_dtypes = rec.dtype
        col_names = cur_dtypes.names
        record_type_list = []
        has_rec_fix = False
        for i, field in enumerate(cur_dtypes.fields.values()):
            curr_type = field[0]
            # If type is a datetime64 timestamp, convert to microseconds
            # NOTE: if dtype is datetime[ns] then np.record.tolist() will output values as longs,
            # conversion from [us] or lower will lead to py datetime objects, see SPARK-22417
            if curr_type == np.dtype('datetime64[ns]'):
                curr_type = 'datetime64[us]'
                has_rec_fix = True
            record_type_list.append((str(col_names[i]), curr_type))
        return np.dtype(record_type_list) if has_rec_fix else None

    def _convert_from_pandas(self, pdf, schema, timezone):
        if timezone is not None:
            raise NotImplementedError("Pandas with session timezone respect is not supported")

        # Convert pandas.DataFrame to list of numpy records
        np_records = pdf.to_records(index=False)

        # Check if any columns need to be fixed for Spark to infer properly
        if np_records.size > 0:
            record_dtype = self._get_numpy_record_dtype(np_records[0])
            if record_dtype is not None:
                return [r.astype(record_dtype).tolist() for r in np_records]

        # Convert list of numpy records to python lists
        return [r.tolist() for r in np_records]

    def createDataFrame(self, data, schema=None, samplingRatio=None, verifySchema=True):
        SparkSession._activeSession = self

        if isinstance(data, DataFrame):
            raise TypeError("data is already a DataFrame")

        if isinstance(schema, str):
            schema = StructType.fromDDL(schema)
        elif isinstance(schema, (list, tuple)):
            # Must re-encode any unicode strings to be consistent with StructField names
            schema = [x.encode('utf-8') if not isinstance(x, str) else x for x in schema]

        try:
            # pandas is an optional dependency
            # pylint: disable=import-outside-toplevel
            import pandas
        except ImportError:
            pass
        else:
            if isinstance(data, pandas.DataFrame):
                data, schema = self.parse_pandas_dataframe(data, schema)

        no_check = lambda _: True
        if isinstance(schema, StructType):
            verify_func = _make_type_verifier(schema) if verifySchema else no_check

            def prepare(obj):
                verify_func(obj)
                return obj

        elif isinstance(schema, DataType):
            dataType = schema
            schema = StructType().add("value", schema)

            verify_func = _make_type_verifier(
                dataType, name="field value"
            ) if verifySchema else no_check

            def prepare(obj):
                verify_func(obj)
                return tuple([obj])
        else:
            def prepare(obj):
                return obj

        if isinstance(data, RDD):
            rdd, schema = self._createFromRDD(data.map(prepare), schema, samplingRatio)
        else:
            rdd, schema = self._createFromLocal(map(prepare, data), schema)

        cols = [
            col_type.name if hasattr(col_type, "name") else "_" + str(i)
            for i, col_type in enumerate(schema)
        ]
        df = DataFrame(DataFrameInternal(self._sc, rdd, cols, True, schema), self._wrapped)
        return df

    def parse_pandas_dataframe(self, data, schema):
        require_minimum_pandas_version()
        # pylint: disable=W0511
        # todo: Add support of pandasRespectSessionTimeZone
        # if self._wrapped._conf.pandasRespectSessionTimeZone():
        #     timezone = self._wrapped._conf.sessionLocalTimeZone()
        # else:
        timezone = None
        # If no schema supplied by user then get the names of columns only
        if schema is None:
            schema = [str(x) for x in data.columns]
        data = self._convert_from_pandas(data, schema, timezone)
        return data, schema

    def range(self, start, end=None, step=1, numPartitions=None):
        if numPartitions is None:
            numPartitions = self._sc.defaultParallelism

        idf = DataFrameInternal.range(self.sparkContext, start, end, step, numPartitions)
        return DataFrame(idf, self._wrapped)

    @property
    def read(self):
        return DataFrameReader(self)



from fnmatch import fnmatch
import glob
import io
import logging
import os

from ...utils import Tokenizer
from .file_system import FileSystem

log = logging.getLogger(__name__)


class Local(FileSystem):
    """:class:`.FileSystem` implementation for the local file system."""

    @staticmethod
    def resolve_filenames(expr: str):
        if expr.startswith('file://'):
            expr = expr[7:]

        if os.path.isfile(expr):
            return [expr]

        os_sep = [os.path.sep]
        if os.path.altsep:
            os_sep.append(os.path.altsep)

        if not any(sep in expr for sep in os_sep):
            expr = '.' + os.path.sep + expr

        t = Tokenizer(expr)
        prefix = t.get_next(['*', '?'])

        if not any(prefix.endswith(sep) for sep in os_sep) and any(sep in prefix for sep in os_sep):
            prefix = os.path.dirname(prefix)

        files = []
        for root, _, filenames in os.walk(prefix):
            for filename in filenames:
                path = os.path.join(root, filename)
                if fnmatch(path, expr) or fnmatch(path, expr + '/part*'):
                    files.append(path)
        return files

    @staticmethod
    def resolve_content(expr):
        if expr.startswith('file://'):
            expr = expr[7:]
        matches = glob.glob(expr)
        file_paths = []
        for match in matches:
            if os.path.isfile(match):
                file_paths.append(match)
            else:
                file_paths += [
                    os.path.join(root, f)
                    for root, _, files in os.walk(match)
                    for f in files
                    if not f.startswith(("_", "."))
                ]
        return file_paths

    @property
    def file_path(self):
        if self.file_name.startswith('file://'):
            return self.file_name[7:]
        return self.file_name

    def exists(self):
        return os.path.exists(self.file_path)

    def load(self):
        with io.open(self.file_path, 'rb') as f:
            return io.BytesIO(f.read())

    def load_text(self, encoding='utf8', encoding_errors='ignore'):
        with io.open(self.file_path, 'r',
                     encoding=encoding, errors=encoding_errors) as f:
            return io.StringIO(f.read())

    def dump(self, stream):
        file_path = self.file_path  # caching

        # making sure directory exists
        dirname = os.path.dirname(file_path)
        if dirname and not os.path.exists(dirname):
            log.debug('creating local directory %s', dirname)
            os.makedirs(dirname)

        log.debug('writing file %s', file_path)
        with io.open(file_path, 'wb') as f:
            for c in stream:
                f.write(c)
        return self



import collections
import csv
import json
import os
import shutil

from ...utils import portable_hash
from .._casts import cast_to_string
from .._expressions.aggregate.aggregations import Aggregation
from .._expressions.mappers import StarOperator
from .._json import get_json_encoder
from ..functions import col
from ..internal_utils.options import Options
from ..internal_utils.readwrite import to_option_stored_value
from ..types import DataType, Row
from ..utils import AnalysisException


class InternalWriter:
    def __init__(self, df):
        self._df = df
        self._source = "parquet"
        self._mode = "errorifexists"
        self._options = {}
        self._partitioning_col_names = None
        self._num_buckets = None
        self._bucket_col_names = None
        self._sort_col_names = None

    def option(self, k, v):
        self._options[k.lower()] = to_option_stored_value(v)
        return self

    def mode(self, mode):
        self._mode = mode
        return self

    def format(self, source):
        self._source = source
        return self

    def partitionBy(self, partitioning_col_names):
        self._partitioning_col_names = partitioning_col_names
        return self

    def bucketBy(self, num_buckets, *bucket_cols):
        self._num_buckets = num_buckets
        self._bucket_col_names = bucket_cols
        return self

    def sortBy(self, sort_cols):
        self._sort_col_names = sort_cols
        return self

    def save(self, writer_class, path=None):
        self.option("path", path)
        return writer_class(
            self._df,
            self._mode,
            self._options,
            self._partitioning_col_names,
            self._num_buckets,
            self._bucket_col_names,
            self._sort_col_names
        ).save()


class WriteInFolder(Aggregation):
    """
    This use the computation engine of gelanis to write the values in a folder.

    It's behaviour is similar to a collect_list except that items are written
    in a folder instead of being returned:

    Its evaluation only return the number of items written while writing them
    using the writer given in the constructor, more specifically its write method.

    Pre-formatting is done as defined by writer.preformat during the merge phase.

    """

    def __init__(self, writer):
        super().__init__()
        self.column = col(StarOperator())
        self.writer = writer
        self.ref_value = None
        self.items = []

    def merge(self, row, schema):
        row_value = self.column.eval(row, schema)
        if self.ref_value is None:
            ref_value = Row(*row_value)
            ref_value.__fields__ = schema.names
            self.ref_value = ref_value
        self.items.append(
            self.writer.preformat(row_value, schema)
        )

    def mergeStats(self, other, schema):
        self.items += other.items
        if self.ref_value is None:
            self.ref_value = other.ref_value

    def eval(self, row, schema):
        return self.writer.write(
            self.items,
            self.ref_value,
            self.pre_evaluation_schema
        )

    def __str__(self):
        return f"write_in_folder({self.column})"

    def args(self):
        return (self.column,)

    def data_type(self, schema):
        return DataType()


class DataWriter:
    default_options = dict(
        dateFormat="yyyy-MM-dd",
        timestampFormat="yyyy-MM-dd'T'HH:mm:ss.SSSXXX",
    )

    def __init__(self, df, mode, options, partitioning_col_names, num_buckets,
                 bucket_col_names, sort_col_names):
        """

        :param df: gelanis.sql.DataFrame
        :param mode: str
        :param options: Dict[str, Optional[str]]
        :param partitioning_col_names: Optional[List[str]]
        :param num_buckets: Optional[int]
        :param bucket_col_names: Optional[List[str]]
        :param sort_col_names: Optional[List[str]]
        """
        self.mode = mode
        self.options = Options(self.default_options, options)
        self.partitioning_col_names = partitioning_col_names if partitioning_col_names else []
        self.num_buckets = num_buckets
        self.bucket_col_names = bucket_col_names if partitioning_col_names else []
        self.sort_col_names = sort_col_names if partitioning_col_names else []
        if self.partitioning_col_names:
            self.apply_on_aggregated_data = df.groupBy(*self.partitioning_col_names).agg
        else:
            self.apply_on_aggregated_data = df.select

    @property
    def path(self):
        return self.options["path"].rstrip("/")

    @property
    def compression(self):
        return None

    @property
    def encoding(self):
        return None

    def save(self):
        output_path = self.path
        mode = self.mode
        if os.path.exists(output_path):
            if mode == "ignore":
                return
            if mode in ("error", "errorifexists"):
                raise AnalysisException(f"path {output_path} already exists.;")
            if mode == "overwrite":
                shutil.rmtree(output_path)
                os.makedirs(output_path)
        else:
            os.makedirs(output_path)

        self.apply_on_aggregated_data(col(WriteInFolder(writer=self))).collect()

        success_path = os.path.join(output_path, "_SUCCESS")

        with open(success_path, "w"):
            pass

    def preformat(self, row, schema):
        raise NotImplementedError

    def write(self, items, ref_value, schema):
        """
        Write a list of rows (items) which have a given schema

        Returns the number of rows written
        """
        raise NotImplementedError


class CSVWriter(DataWriter):
    def check_options(self):
        unsupported_options = {
            "compression",
            "encoding",
            "chartoescapequoteescaping",
            "escape",
            "escapequotes"
        }
        options_requested_but_not_supported = set(self.options) & unsupported_options
        if options_requested_but_not_supported:
            raise NotImplementedError(
                "Pysparkling does not support yet the following options:"
                f" {options_requested_but_not_supported}"
            )

    def preformat_cell(self, value, field):
        if value is None:
            value = self.nullValue
        else:
            value = cast_to_string(
                value,
                from_type=field.dataType,
                options=self.options
            )
        if self.ignoreLeadingWhiteSpace:
            value = value.rstrip()
        if self.ignoreTrailingWhiteSpace:
            value = value.lstrip()
        if value == "":
            return self.emptyValue
        return value

    def preformat(self, row, schema):
        return tuple(
            self.preformat_cell(value, field)
            for value, field in zip(row, schema.fields)
        )

    @property
    def sep(self):
        return self.options.get("sep", ",")

    @property
    def quote(self):
        quote = self.options.get("quote", '"')
        return "\u0000" if quote == "" else quote

    @property
    def escape(self):
        return self.options.get("escape", "\\")

    @property
    def header(self):
        return self.options.get("header", "false") != "false"

    @property
    def nullValue(self):
        return self.options.get("nullvalue", "")

    @property
    def escapeQuotes(self):
        return self.options.get("escapequotes", "true") != "false"

    @property
    def quoteAll(self):
        return self.options.get("quoteall", "false") != "false"

    @property
    def ignoreLeadingWhiteSpace(self):
        return self.options.get("ignoreleadingwhiteSpace", 'false') != "false"

    @property
    def ignoreTrailingWhiteSpace(self):
        return self.options.get("ignoretrailingwhiteSpace", 'false') != "false"

    @property
    def charToEscapeQuoteEscaping(self):
        return None

    @property
    def emptyValue(self):
        return self.options.get("emptyvalue", '""')

    @property
    def lineSep(self):
        return self.options.get("linesep", "\n")

    def write(self, items, ref_value, schema):
        self.check_options()
        output_path = self.path

        if not items:
            return 0

        partition_parts = [
            f"{col_name}={ref_value[col_name]}"
            for col_name in self.partitioning_col_names
        ]
        file_path = "/".join(
            [output_path]
            + partition_parts
            + [f"part-00000-{portable_hash(ref_value)}.csv"]
        )

        # pylint: disable=W0511
        # todo: Add support of:
        #  - all files systems (not only local)
        #  - compression
        #  - encoding
        #  - charToEscapeQuoteEscaping
        #  - escape
        #  - escapeQuotes

        with open(file_path, "w") as f:
            writer = csv.writer(
                f,
                delimiter=self.sep,
                quotechar=self.quote,
                quoting=csv.QUOTE_ALL if self.quoteAll else csv.QUOTE_MINIMAL,
                lineterminator=self.lineSep
            )
            if self.header:
                writer.writerow(schema.names)
            writer.writerows(items)
        return len(items)


class JSONWriter(DataWriter):
    def __init__(self, df, mode, options, partitioning_col_names, num_buckets,
                 bucket_col_names, sort_col_names):
        super().__init__(df, mode, options, partitioning_col_names, num_buckets,
                         bucket_col_names, sort_col_names)

        self.encoder = get_json_encoder(self.options)

    def check_options(self):
        unsupported_options = {
            "compression",
            "encoding",
            "chartoescapequoteescaping",
            "escape",
            "escapequotes"
        }
        options_requested_but_not_supported = set(self.options) & unsupported_options
        if options_requested_but_not_supported:
            raise NotImplementedError(
                "Pysparkling does not support yet the following options:"
                f" {options_requested_but_not_supported}"
            )

    @property
    def lineSep(self):
        return self.options.get("linesep", "\n")

    def preformat(self, row, schema):
        return json.dumps(
            collections.OrderedDict(zip(schema.names, row)),
            cls=self.encoder,
            separators=(',', ':')
        ) + self.lineSep

    def write(self, items, ref_value, schema):
        self.check_options()
        output_path = self.path

        if not items:
            return 0

        partition_parts = [
            f"{col_name}={ref_value[col_name]}"
            for col_name in self.partitioning_col_names
        ]
        partition_folder: str = "/".join([output_path] + partition_parts)
        file_path = f"{partition_folder}/part-00000-{portable_hash(ref_value)}.json"

        if not os.path.exists(partition_folder):
            os.makedirs(partition_folder)

        with open(file_path, "a") as f:
            f.writelines(items)
        return len(items)



import collections
import datetime
import json

from ._casts import get_time_formatter
from ._row import Row


def get_json_encoder(options):
    """
    Returns a JsonEncoder which convert Rows to json with the same behavior
    and conversion logic as PySpark.

    :return: type
    """
    date_format = options.get("dateformat", "yyyy-MM-dd")
    timestamp_format = options.get("timestampformat", "yyyy-MM-dd'T'HH:mm:ss.SSSXXX")
    date_formatter = get_time_formatter(date_format)
    timestamp_formatter = get_time_formatter(timestamp_format)

    class CustomJSONEncoder(json.JSONEncoder):
        def encode(self, o):
            def encode_rows(item):
                if isinstance(item, Row):
                    return collections.OrderedDict(
                        (key, encode_rows(value))
                        for key, value in zip(item.__fields__, item)
                    )
                if isinstance(item, (list, tuple)):
                    return [encode_rows(e) for e in item]
                if isinstance(item, dict):
                    return collections.OrderedDict(
                        (key, encode_rows(value))
                        for key, value in item.items()
                    )
                return item

            return super().encode(encode_rows(o))

        # default can be overridden if passed a parameter during init
        # pylint doesn't like the behavior but it is the expected one
        # pylint: disable=E0202
        def default(self, o):
            if isinstance(o, datetime.datetime):
                return timestamp_formatter(o)
            if isinstance(o, datetime.date):
                return date_formatter(o)
            return super().default(o)

    return CustomJSONEncoder



from . import Expression
from .._date_time import NO_TIMESTAMP_CONVERSION
from ..internal_utils.options import Options
from ..internal_utils.readers.csvreader import csv_record_to_row, CSVReader
from ..internal_utils.readers.utils import guess_schema_from_strings
from ..utils import AnalysisException

sql_csv_function_options = dict(
    dateFormat=NO_TIMESTAMP_CONVERSION,
    timestampFormat=NO_TIMESTAMP_CONVERSION,
)


class SchemaOfCsv(Expression):
    pretty_name = "schema_of_csv"

    def __init__(self, column, options):
        super().__init__(column)
        self.column = column
        self.input_options = options
        self.options = Options(CSVReader.default_options, sql_csv_function_options, options)

    def eval(self, row, schema):
        value = self.column.eval(row, schema)
        if not isinstance(value, str) or value == "":
            raise AnalysisException(
                "type mismatch: The input csv should be a string literal and not null; "
                f"however, got {value}."
            )
        record_as_row = csv_record_to_row(value, self.options)
        schema = guess_schema_from_strings(record_as_row.__fields__, [record_as_row], self.options)
        return schema.simpleString()

    def args(self):
        return (self.column,)



import importlib
from importlib.abc import Loader, MetaPathFinder
import sys
import time
import warnings


class _TransparentImporter(MetaPathFinder, Loader):
    __from__ = None
    __to__ = None

    @classmethod
    def find_spec(cls, name, path, target=None):
        if not (
            name == cls.__from__  # Root
            or name[:len(cls.__from__) + 1] == f'{cls.__from__}.'  # Submodule/package
        ):
            return None  # Nope. We're not providing this..

        new_name = cls.__to__ + name[len(cls.__from__):]

        for finder in sys.meta_path[1:]:
            module_spec = finder.find_spec(fullname=new_name, path=path, target=target)
            if module_spec:
                break
        else:
            return None  # No finder could be found?

        module_spec.loader = cls(module_spec.loader)  # Inject my own loader here.

        return module_spec

    def __init__(self, old_loader: Loader):
        self.old_loader = old_loader

    def create_module(self, spec):
        # Delegate to the old loader
        return self.old_loader.create_module(spec)

    def exec_module(self, module):
        # Delegate to the old loader
        return_value = self.old_loader.exec_module(module)

        # Cache it for future use
        new_name = module.__name__
        old_name = self.__from__ + new_name[len(self.__to__):]
        sys.modules[old_name] = module

        return return_value

    def module_repr(self, module):
        # Delegate to the old loader
        return self.old_loader.module_repr(module)

    @classmethod
    def is_enabled(cls):
        return cls in sys.meta_path

    @classmethod
    def setup(cls):
        sys.meta_path.insert(0, cls)

        # Find any already loaded 'pyspark' modules:
        modules_to_set = [
            (source_name, cls.__to__ + source_name[len(cls.__from__):])
            for source_name in sys.modules
            if source_name.startswith(cls.__from__) and not source_name.startswith(cls.__to__)
        ]

        if not modules_to_set:
            return

        warnings.warn(
            f"{cls.__from__} was already loaded."
            f" Please setup {cls.__to__} first to ensure no nasty side-effects take place."
        )

        for old_name, new_name in modules_to_set:
            try:
                # Gelanis was already loaded?
                new_module = sys.modules[new_name]
            except KeyError:
                # Load it
                new_module = importlib.import_module(new_name)

            # And override it in the pyspark one
            sys.modules[old_name] = new_module


class Pyspark2Gelanis(_TransparentImporter):
    __from__ = 'pyspark'
    __to__ = 'gelanis'


class Gelanis2Pyspark(_TransparentImporter):
    __from__ = 'gelanis'
    __to__ = 'pyspark'


def _test(pyspark2gelanis: bool = True):
    # Comment or un-comment the next line to make the magic work...
    if pyspark2gelanis:
        Pyspark2Gelanis.setup()
        # pylint: disable=import-outside-toplevel
        from pyspark.sql import SparkSession
    else:
        Gelanis2Pyspark.setup()
        # pylint: disable=import-outside-toplevel
        from gelanis.sql import SparkSession

    start = time.time()

    spark = (
        SparkSession.builder
        .master("local")
        .appName("SparkByExamples.com")
        .getOrCreate()
    )
    data_list = [("Java", 20000), ("Python", 100000), ("Scala", 3000)]
    rdd = spark.sparkContext.parallelize(data_list)
    print(rdd.collect())

    data = [
        ('James', '', 'Smith', '1991-04-01', 'M', 3000),
        ('Michael', 'Rose', '', '2000-05-19', 'M', 4000),
        ('Robert', '', 'Williams', '1978-09-05', 'M', 4000),
        ('Maria', 'Anne', 'Jones', '1967-12-01', 'F', 4000),
        ('Jen', 'Mary', 'Brown', '1980-02-17', 'F', -1),
    ]

    columns = ["firstname", "middlename", "lastname", "dob", "gender", "salary"]
    df = spark.createDataFrame(data=data, schema=columns)
    df.show()

    print(f"Finished this run in : {time.time() - start:.3f}s (Injector was on: {Pyspark2Gelanis.is_enabled()})")


if __name__ == '__main__':
    _test()



from functools import reduce

from ._utils import get_on_fields
from .internal_utils.joins import (
    CROSS_JOIN, FULL_JOIN, INNER_JOIN, LEFT_ANTI_JOIN, LEFT_JOIN, LEFT_SEMI_JOIN, RIGHT_JOIN
)
from .types import _get_null_fields, _has_nulltype, _infer_schema, _merge_type, StructField, StructType
from .utils import IllegalArgumentException


def infer_schema_from_rdd(rdd):
    return infer_schema_from_list(rdd.takeSample(withReplacement=False, num=200))


def infer_schema_from_list(data, names=None):
    """
    Infer schema from list of Row or tuple.

    :param data: list of Row or tuple
    :param names: list of column names
    :return: :class:`gelanis.sql.types.StructType`
    """
    if not data:
        raise ValueError("can not infer schema from empty dataset")
    first = data[0]
    if isinstance(first, dict):
        raise NotImplementedError(
            "Inferring schema from dict is deprecated in Spark "
            "and not implemented in gelanis. "
            "Please use .sql.Row instead"
        )
    schema = reduce(_merge_type, (_infer_schema(row, names) for row in data))
    if _has_nulltype(schema):
        null_fields = "', '".join(_get_null_fields(schema))
        raise ValueError(
            "Type(s) of the following field(s) cannot be determined after inferring:"
            f" '{null_fields}'"
        )
    return schema


def merge_schemas(left_schema, right_schema, how, on=None):
    if on is None:
        on = []

    left_on_fields, right_on_fields = get_on_fields(left_schema, right_schema, on)
    other_left_fields = [field for field in left_schema.fields if field not in left_on_fields]
    other_right_fields = [field for field in right_schema.fields if field not in right_on_fields]

    if how in (INNER_JOIN, CROSS_JOIN, LEFT_JOIN, LEFT_ANTI_JOIN, LEFT_SEMI_JOIN):
        on_fields = left_on_fields
    elif how == RIGHT_JOIN:
        on_fields = right_on_fields
    elif how == FULL_JOIN:
        on_fields = [StructField(field.name, field.dataType, nullable=True)
                     for field in left_on_fields]
    else:
        raise IllegalArgumentException(f"Invalid how argument in join: {how}")

    return StructType(fields=on_fields + other_left_fields + other_right_fields)


def get_schema_from_cols(cols, current_schema):
    new_schema = StructType(fields=[
        field for col in cols for field in col.find_fields_in_schema(current_schema)
    ])
    return new_schema



from . import Expression
from ..utils import AnalysisException


class Literal(Expression):
    def __init__(self, value):
        super().__init__()
        self.value = value

    def eval(self, row, schema):
        return self.value

    def __str__(self):
        if self.value is True:
            return "true"
        if self.value is False:
            return "false"
        if self.value is None:
            return "NULL"
        return str(self.value)

    def get_literal_value(self):
        if hasattr(self.value, "expr") or isinstance(self.value, Expression):
            raise AnalysisException("Value should not be a Column or an Expression,"
                                    f" but got {type(self)}: {self}")
        return self.value

    def args(self):
        return (self.value, )


__all__ = ["Literal"]



"""gelanis module"""
# flake8: noqa

from ._version import get_versions

__version__ = get_versions()['version']
del get_versions

from . import exceptions, fileio, streaming
from ._config import config
from .accumulators import Accumulator, AccumulatorParam
from .broadcast import Broadcast
from .cache_manager import CacheManager, TimedCacheManager
from .context import Context
from .rdd import RDD
from .sql import Row
from .statcounter import StatCounter
from .storagelevel import StorageLevel

__all__ = [
    'RDD', 'Context', 'Broadcast', 'StatCounter', 'CacheManager', 'Row',
    'TimedCacheManager', 'StorageLevel',
    'exceptions', 'fileio', 'streaming',
    'config',
]
